from typing import TYPE_CHECKING, Callable, List, Optional, Union

import audobject
import torch
from torchvision import transforms as T

from .abstract_transform import AbstractTransform


if TYPE_CHECKING:
    from autrainer.datasets import AbstractDataset  # pragma: no cover


class SmartCompose(T.Compose, audobject.Object):
    def __init__(
        self,
        transforms: List[Union[T.Compose, AbstractTransform, Callable]],
        **kwargs,
    ) -> None:
        """SmartCompose wrapper for torchvision.transforms.Compose, which
        allows for simple composition of transforms by adding them together.
        Transforms are automatically sorted by their order attribute if present.

        Additionally, SmartCompose allows for the specification of a
        collate function to be used in a DataLoader, the last collate function
        specified will be used.

        Args:
            transforms: List of transforms to compose.
            **kwargs: Additional keyword arguments to pass to
                torchvision.transforms.Compose.
        """
        super().__init__(transforms, **kwargs)
        self._sort()

    def __add__(
        self,
        other: Union[T.Compose, AbstractTransform, Callable, List[Callable]],
    ) -> "SmartCompose":
        """Add another transform to the composition. Transforms are
        automatically sorted by their order attribute if present.

        Args:
            other: Transform to add to the composition.

        Raises:
            TypeError: If the addition is not a valid type.
                Supported types are 'torchvision.transforms.Compose',
                'AbstractTransform', 'Callable', or 'List[Callable]'.

        Returns:
            New SmartCompose object with the added transform.
        """
        if other is None:
            return self
        elif isinstance(other, T.Compose):
            return SmartCompose(
                self.transforms + other.transforms,
            )
        elif isinstance(other, (AbstractTransform, Callable)):
            return SmartCompose(self.transforms + [other])
        elif isinstance(other, List):
            return SmartCompose(self.transforms + other)
        else:
            raise TypeError(
                "SmartCompose addition must be a "
                "'torchvision.transforms.Compose', 'AbstractTransform', "
                f"'Callable', or 'List[Callable]' got '{type(other)}'."
            )

    def _sort(self):
        """Sort the transforms by their order attribute if present."""
        self.transforms.sort(key=lambda x: getattr(x, "order", 0))

    def get_collate_fn(self, data: "AbstractDataset") -> Optional[Callable]:
        """Get the collate function. If no collate function is present in
        the transforms, None is returned.
        If multiple collate functions are present, the last one is used.

        Args:
            data: Dataset to get the collate function for.

        Returns:
            Collate function.
        """
        collate_fn = None
        for t in self.transforms:
            if fn := getattr(t, "get_collate_fn", None):
                collate_fn = fn
        if collate_fn is not None:
            return collate_fn(data)

    def __call__(self, x: torch.Tensor, index: int) -> torch.Tensor:
        """Apply the transforms to the input tensor.

        Args:
            x: Input tensor.
            index: Dataset index of the input tensor.

        Returns:
            Transformed tensor.
        """
        for t in self.transforms:
            if "index" in t.__call__.__annotations__.keys():
                x = t(x, index=index)
            else:
                x = t(x)
        return x



from .constants import (
    CONFIG_FOLDERS,
    INVALID_AGGREGATIONS,
    NAMING_CONVENTION,
    VALID_AGGREGATIONS,
)


__all__ = [
    "CONFIG_FOLDERS",
    "INVALID_AGGREGATIONS",
    "NAMING_CONVENTION",
    "VALID_AGGREGATIONS",
]



import torch
from transformers import WhisperModel

from .abstract_model import AbstractModel
from .ffnn import FFNN


class WhisperBackbone(AbstractModel):
    def __init__(self, model_name: str) -> None:
        self.model_name = model_name
        model = WhisperModel.from_pretrained(self.model_name)
        super().__init__(output_dim=model.config.hidden_size)
        self.model = model
        self.register_buffer(
            "decoder_input_ids",
            torch.tensor([1, 1]) * self.model.config.decoder_start_token_id,
        )

    def embeddings(self, x: torch.Tensor) -> torch.Tensor:
        decoder_input_ids = torch.stack([self.decoder_input_ids] * x.shape[0])
        x = self.model(x, decoder_input_ids=decoder_input_ids)[
            "last_hidden_state"
        ][:, 0, :]
        return x

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        return self.embeddings(x)


class WhisperFFNN(AbstractModel):
    def __init__(
        self,
        output_dim: int,
        model_name: str,
        hidden_size: int,
        num_layers: int = 2,
        sigmoid: bool = False,
        softmax: bool = False,
        dropout: float = 0.5,
    ) -> None:
        """Whisper model with FFNN frontend adapted for audio classification.
        For more information, see:
        https://doi.org/10.48550/arXiv.2212.04356

        Args:
            model_name: Name of the model loaded from Huggingface.
            hidden_size: Hidden size of the FFNN.
            output_dim: Output dimension of the FFNN.
            num_layers: Number of layers of the FFNN. Defaults to 2.
            sigmoid: Whether to apply sigmoid activation. Defaults to False.
            softmax: Whether to apply softmax activation. Defaults to False.
            dropout: Dropout rate. Defaults to 0.5.
        """
        super().__init__(output_dim)
        self.model_name = model_name
        self.hidden_size = hidden_size
        self.num_layers = num_layers
        self.sigmoid = sigmoid
        self.softmax = softmax
        self.dropout = dropout
        self.backbone = WhisperBackbone(model_name=model_name)
        self.frontend = FFNN(
            input_size=self.backbone.output_dim,
            hidden_size=hidden_size,
            output_dim=output_dim,
            num_layers=num_layers,
            sigmoid=sigmoid,
            softmax=softmax,
            dropout=dropout,
        )

    def embeddings(self, x: torch.Tensor) -> torch.Tensor:
        return self.backbone(x)

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        return self.frontend(self.embeddings(x))



from abc import abstractmethod
from typing import Optional

import torch

from autrainer.transforms import AbstractTransform


class AbstractAugmentation(AbstractTransform):
    def __init__(
        self,
        order: int = 0,
        p: float = 1.0,
        generator_seed: Optional[int] = None,
        **kwargs,
    ) -> None:
        """Abstract class for an augmentation.

        Args:
            order: The order of the augmentation in the transformation pipeline.
                Defaults to 0.
            p: The probability of applying the augmentation. Defaults to 1.0.
            generator_seed: The initial seed for the internal random number
                generator drawing the probability. If None, the generator is
                not seeded. Defaults to None.
            kwargs: Additional keyword arguments to store in the object.

        Raises:
            ValueError: If p is not in the range [0, 1].
        """
        super().__init__(order, **kwargs)
        if p < 0 or p > 1:
            raise ValueError("p must be in the range [0, 1]")
        self.p = p
        self.generator_seed = generator_seed
        self.g = torch.Generator()
        if self.generator_seed is not None:
            self.g.manual_seed(self.generator_seed)

    def __call__(self, x: torch.Tensor, index: int = None) -> torch.Tensor:
        """Call the augmentation apply method with probability p.

        Args:
            x: The input tensor.
            index: The index of the input tensor in the dataset.
                Defaults to None.

        Returns:
            The augmented tensor if the probability is less than p, otherwise
            the input tensor.
        """
        probability = torch.rand(1, generator=self.g).item()
        return self.apply(x, index) if probability < self.p else x

    @abstractmethod
    def apply(self, x: torch.Tensor, index: int = None) -> torch.Tensor:
        """Apply the augmentation to the input tensor.

        Apply is called with probability p.

        Args:
            x: The input tensor.
            index: The index of the input tensor in the dataset.
                Defaults to None.

        Returns:
            The augmented tensor.
        """



from typing import Tuple

import torch

import autrainer


class SAM(torch.optim.Optimizer):
    def __init__(
        self,
        params: torch.nn.Module,
        base_optimizer: str,
        rho: float = 0.05,
        adaptive: bool = False,
        **kwargs,
    ):
        """Sharpness Aware Minimization (SAM) optimizer.

        This implementation is adapted from the following repository:
        https://github.com/davda54/sam

        For more information, see:
        https://arxiv.org/abs/2010.01412

        Args:
            params: Model parameters.
            base_optimizer: Underlying optimizer performing the sharpness-aware
                update, specified as a relative import path.
            rho: Size of the neighborhood for computing the max loss.
                Defaults to 0.05.
            adaptive: Whether to use an experimental implementation of
                element-wise Adaptive SAM. Defaults to False.
            **kwargs: Additional arguments passed to the underlying optimizer.
        """
        if rho < 0.0:
            raise ValueError(f"Rho '{rho}' should be non-negative.")

        defaults = dict(rho=rho, adaptive=adaptive, **kwargs)
        super(SAM, self).__init__(params, defaults)

        self.base_optimizer = autrainer.instantiate(
            config={
                "_target_": base_optimizer,
                "params": self.param_groups,
                **kwargs,
            },
            instance_of=torch.optim.Optimizer,
        )
        self.param_groups = self.base_optimizer.param_groups
        self.defaults.update(self.base_optimizer.defaults)

    @torch.no_grad()
    def first_step(self, zero_grad=False):
        grad_norm = self._grad_norm()
        for group in self.param_groups:
            scale = group["rho"] / (grad_norm + 1e-12)
            for p in group["params"]:
                if p.grad is None:
                    continue
                self.state[p]["old_p"] = p.data.clone()
                e_w = (
                    (torch.pow(p, 2) if group["adaptive"] else 1.0)
                    * p.grad
                    * scale.to(p)
                )
                p.add_(e_w)  # climb to the local maximum "w + e(w)"

        if zero_grad:
            self.zero_grad()

    # @torch.no_grad()
    # def first_step_uphill(self, zero_grad=False):
    #     grad_norm = self._grad_norm()
    #     for group in self.param_groups:
    #         scale = group["rho"] / (grad_norm + 1e-12)

    #         for p in group["params"]:
    #             if p.grad is None: continue
    #             self.state[p]["old_p"] = p.data.clone()
    #             e_w = (torch.pow(p, 2) if group["adaptive"] else 1.0) * p.grad * scale.to(p)
    #             p.add_(e_w)  # climb to the local maximum "w + e(w)"

    #     if zero_grad: self.zero_grad()

    @torch.no_grad()
    def second_step(self, zero_grad=False):
        for group in self.param_groups:
            for p in group["params"]:
                if p.grad is None:
                    continue
                # get back to "w" from "w + e(w)"
                p.data = self.state[p]["old_p"]

        self.base_optimizer.step()  # do the actual "sharpness-aware" update

        if zero_grad:
            self.zero_grad()

    @torch.no_grad()
    def step(self, closure=None):
        assert (
            closure is not None
        ), "Sharpness Aware Minimization requires closure, but it was not provided"
        # the closure should do a full forward-backward pass
        closure = torch.enable_grad()(closure)

        self.first_step(zero_grad=True)
        closure()
        self.second_step()

    def _grad_norm(self):
        # put everything on the same device, in case of model parallelism
        shared_device = self.param_groups[0]["params"][0].device
        norm = torch.norm(
            torch.stack(
                [
                    ((torch.abs(p) if group["adaptive"] else 1.0) * p.grad)
                    .norm(p=2)
                    .to(shared_device)
                    for group in self.param_groups
                    for p in group["params"]
                    if p.grad is not None
                ]
            ),
            p=2,
        )
        return norm

    def load_state_dict(self, state_dict):
        super().load_state_dict(state_dict)
        self.base_optimizer.param_groups = self.param_groups

    # ? Training Tips functions from the Github repo ref: https://github.com/davda54/sam
    # TODO: Check if Batch Norm Tip needs to be applied
    def custom_step(
        self,
        model: torch.nn.Module,
        data: torch.Tensor,
        target: torch.Tensor,
        criterion: torch.nn.Module,
    ) -> Tuple[float, torch.Tensor]:
        output = model(data)
        loss = criterion(output, target)
        loss.backward()
        self.first_step(zero_grad=True)

        output = model(data)
        loss = criterion(output, target)
        loss.backward()
        self.second_step(zero_grad=True)
        _loss = loss.item()
        return _loss, output.detach()



from dataclasses import dataclass
import glob
import os
from typing import List, Optional, Union

from autrainer.core.scripts.abstract_script import MockParser

from .abstract_postprocess_script import (
    AbstractPostprocessArgs,
    AbstractPostprocessScript,
)
from .utils import catch_cli_errors


@dataclass
class DeleteStatesArgs(AbstractPostprocessArgs):
    keep_best: bool
    keep_runs: Optional[List[str]]
    keep_iterations: Optional[List[int]]


class DeleteStatesScript(AbstractPostprocessScript):
    def __init__(self) -> None:
        super().__init__(
            "rm-states",
            "Delete states (.pt files) from an experiment.",
            epilog="Example: autrainer rm-states results default",
            dataclass=DeleteStatesArgs,
        )

    def add_arguments(self) -> None:
        super().add_arguments()
        self.parser.add_argument(
            "-b",
            "--keep-best",
            action="store_false",
            default=True,
            help="Keep best states. Defaults to True.",
        )
        self.parser.add_argument(
            "-r",
            "--keep-runs",
            nargs="+",
            metavar="R",
            default=None,
            type=str,
            help="Runs to keep.",
        )
        self.parser.add_argument(
            "-i",
            "--keep-iterations",
            nargs="+",
            metavar="I",
            default=None,
            type=int,
            help="Iterations to keep.",
        )

    def main(self, args: DeleteStatesArgs) -> None:
        super().main(args)
        self._clean_states(args)

    def _clean_states(self, args: DeleteStatesArgs) -> None:
        state_paths = self._collect_state_paths(args)
        if not state_paths:
            return
        if args.keep_best:
            state_paths = self._filter_best_states(state_paths)
        if args.keep_runs:
            state_paths = self._filter_runs(state_paths, args.keep_runs)
        if args.keep_iterations:
            state_paths = self._filter_iterations(
                state_paths,
                args.keep_iterations,
            )
        if self._confirm_deletion(state_paths):
            self._delete_states(state_paths)

    def _collect_state_paths(self, args: DeleteStatesArgs) -> List[str]:
        pattern = "**/*.pt"
        path = os.path.join(args.results_dir, args.experiment_id, "training")
        return [
            f for f in glob.glob(os.path.join(path, pattern), recursive=True)
        ]

    def _filter_best_states(self, state_paths: List[str]) -> List[str]:
        return [f for f in state_paths if "_best" != self._get_iter_dir(f)]

    def _filter_runs(
        self,
        state_paths: List[str],
        runs: List[str],
    ) -> List[str]:
        for run in runs:
            state_paths = [
                f for f in state_paths if run != self._get_run_name(f)
            ]
        return state_paths

    def _filter_iterations(
        self,
        state_paths: List[str],
        iterations: List[int],
    ) -> List[str]:
        for iteration in iterations:
            state_paths = [
                f for f in state_paths if iteration != self._get_iter_num(f)
            ]
        return state_paths

    def _get_iter_dir(self, state_path: str) -> str:
        return os.path.basename(os.path.dirname(state_path))

    def _get_run_name(self, state_path: str) -> str:
        return os.path.basename(os.path.dirname(os.path.dirname(state_path)))

    def _get_iter_num(self, state_path: str) -> Union[int, str]:
        iter_dir = self._get_iter_dir(state_path)
        if iter_dir in ["_best", "_initial"]:
            return iter_dir
        return int(iter_dir.split("_")[-1])

    def _confirm_deletion(self, state_paths: List[str]) -> bool:
        print("The following states will be deleted:")
        print("\n".join(state_paths))
        print()
        response = input(
            "Are you sure you want to delete "
            f"{len(state_paths)} states? (y/n): "
        ).lower()
        return response in ["y", "yes"]

    def _delete_states(self, state_paths: List[str]) -> None:
        for state_path in state_paths:
            try:
                os.remove(state_path)
                print(f"Deleted: {state_path}")
            except Exception as e:
                print(f"Error deleting {state_path}: {e}")  # pragma: no cover


@catch_cli_errors
def rm_states(
    results_dir: str,
    experiment_id: str,
    keep_best: bool = True,
    keep_runs: Optional[List[str]] = None,
    keep_iterations: Optional[List[int]] = None,
) -> None:
    """Delete states (.pt files) from an experiment.

    If called in a notebook, the function will not raise an error and print
    the error message instead.

    Args:
        results_dir: Path to grid search results directory.
        experiment_id: ID of experiment to postprocess.
        keep_best: Keep best states. Defaults to True.
        keep_runs: Runs to keep. Defaults to None.
        keep_iterations: Iterations to keep. Defaults to None.

    Raises:
        CommandLineError: If the results directory or experiment ID dont exist.
    """
    script = DeleteStatesScript()
    script.parser = MockParser()
    script.main(
        DeleteStatesArgs(
            results_dir,
            experiment_id,
            keep_best,
            keep_runs,
            keep_iterations,
        )
    )



import os
from typing import Dict, Optional

import torch
import torch.nn.common_types


def format_results(
    results: Dict[str, float],
    results_type: str,
    training_type: str,
    iteration: Optional[int] = None,
) -> str:
    s = f"{results_type} results"
    s += f" at {training_type} {iteration}" if iteration else ""
    s += ":\n"
    max_key_len = max([len(k) for k in results.keys()])
    s += "\n".join(
        [f"{(k+':').ljust(max_key_len+1)} {v:.4f}" for k, v in results.items()]
    )
    return s


def load_pretrained_model_state(
    model: torch.nn.Module,
    state_dict: Dict[str, torch.Tensor],
    skip_last_layer: bool = True,
) -> None:
    """Load a pretrained model state dict for a model.

    Args:
        model: Model to load the state dict into.
        state_dict: State dict to load into the model.
        skip_last_layer: Whether to skip loading the state dict of the
            last linear or convolutional layer. Defaults to True.
    Raises:
        RuntimeError: If the shapes of the model and state dict do not match.
    """
    last_layer = None
    strict = True
    for name, layer in model.named_modules():
        if isinstance(layer, (torch.nn.Linear, torch.nn.modules.conv._ConvNd)):
            last_layer = name

    if last_layer is not None and skip_last_layer:
        state_dict.pop(last_layer + ".weight", None)
        state_dict.pop(last_layer + ".bias", None)
        strict = False

    model.load_state_dict(state_dict, strict=strict)


def load_pretrained_optim_state(
    optim: torch.optim.Optimizer,
    state_dict: Dict[str, torch.Tensor],
    skip_last_layer: bool = True,
) -> None:
    """Load a pretrained optimizer state dict for an optimizer.

    Args:
        optim: Optimizer to load the state dict into.
        state_dict: State dict to load into the optimizer.
        skip_last_layer: Whether to skip loading the state dict of the
            last layer. Defaults to True.
    """
    if skip_last_layer:
        keys = list(state_dict["state"].keys())
        state_dict["state"].pop(keys[-1], None)  # bias
        state_dict["state"].pop(keys[-2], None)  # weight

    optim.load_state_dict(state_dict)


def load_checkpoint(checkpoint: str) -> Dict[str, torch.Tensor]:
    """Load a checkpoint state dict from a file.

    Args:
        checkpoint: Path to the checkpoint file.

    Raises:
        FileNotFoundError: If the checkpoint file does not exist.

    Returns:
        The model state dict.
    """
    # TODO: Support Hugging Face Checkpoints in the future.
    if not os.path.isfile(checkpoint):
        raise FileNotFoundError(f"Checkpoint file not found: {checkpoint}")
    return torch.load(checkpoint)



NAMING_CONVENTION = [
    "dataset",
    "model",
    "optimizer",
    "learning_rate",
    "batch_size",
    "training_type",
    "iterations",
    "scheduler",
    "augmentation",
    "seed",
]
"""Naming convention of runs."""

INVALID_AGGREGATIONS = [
    "training_type",
]
"""Invalid aggregations for postprocessing."""

VALID_AGGREGATIONS = list(set(NAMING_CONVENTION) - set(INVALID_AGGREGATIONS))
"""Valid aggregations for postprocessing."""

CONFIG_FOLDERS = [
    "augmentation",
    "dataset",
    "model",
    "optimizer",
    "plotting",
    "preprocessing",
    "scheduler",
    "sharpness",
]
"""Configuration folders for Hydra configurations."""



from typing import Optional

import hydra
from omegaconf import DictConfig, OmegaConf

import autrainer
from autrainer.core.scripts.abstract_script import MockParser

from .abstract_preprocess_script import (
    AbstractPreprocessScript,
    PreprocessArgs,
)
from .utils import (
    add_hydra_args_to_sys,
    catch_cli_errors,
    run_hydra_cmd,
    running_in_notebook,
)


class FetchScript(AbstractPreprocessScript):
    def __init__(self) -> None:
        super().__init__(
            "fetch",
            (
                "Fetch the datasets and models specified in a "
                "training configuration (Hydra)."
            ),
            extended_description=(
                "For more information on Hydra's command line line flags, see:\n"
                "https://hydra.cc/docs/advanced/hydra-command-line-flags/."
            ),
            epilog="Example: autrainer fetch -cn config.yaml",
            dataclass=PreprocessArgs,
            unknown_args=True,
        )

    def main(self, args: PreprocessArgs) -> None:
        self._override_launcher(args)
        self.datasets = {}
        self.models = {}

        @autrainer.main("config")
        def main(cfg: DictConfig) -> None:
            if not self._id_in_dict(self.datasets, cfg.dataset.id):
                self.datasets[cfg.dataset.id] = OmegaConf.to_container(
                    cfg.dataset
                )

            if not self._id_in_dict(self.models, cfg.model.id):
                self.models[cfg.model.id] = OmegaConf.to_container(cfg.model)

        main()
        self._download_datasets()
        self._download_models()
        self._clean_up()

    def _download_datasets(self) -> None:
        print("Fetching datasets...")
        for name, dataset in self.datasets.items():
            print(f" - {name}")
            if not dataset.get("path", None):
                continue
            hydra.utils.instantiate(
                config={
                    "_target_": dataset["_target_"] + ".download",
                    "path": dataset["path"],
                },
            )  # pragma: no cover

    def _download_models(self) -> None:
        print("Fetching models...")
        for name, model in self.models.items():
            print(f" - {name}")
            model.pop("transform", None)
            autrainer.instantiate(config=model, output_dim=10)


@catch_cli_errors
def fetch(
    override_kwargs: Optional[dict] = None,
    cfg_launcher: bool = False,
    config_name: str = "config",
    config_path: Optional[str] = None,
) -> None:
    """Fetch the datasets and models specified in a training configuration.

    Args:
        override_kwargs: Additional Hydra override arguments to pass to the
            train script.
        cfg_launcher: Use the launcher specified in the configuration instead
            of the Hydra basic launcher. Defaults to False.
        config_name: The name of the config (usually the file name without the
            .yaml extension). Defaults to "config".
        config_path: The config path, a directory where Hydra will search for
            config files. If config_path is None no directory is added to the
            search path. Defaults to None.
    """
    if running_in_notebook():
        run_hydra_cmd(
            "fetch -l" if cfg_launcher else "fetch",
            override_kwargs,
            config_name,
            config_path,
        )

    else:
        add_hydra_args_to_sys(override_kwargs, config_name, config_path)
        script = FetchScript()
        script.parser = MockParser()
        script.main(PreprocessArgs(cfg_launcher))



import os
from typing import Dict, List, Optional, Tuple, Union

from omegaconf import DictConfig
import pandas as pd
import torch

from autrainer.transforms import AbstractTransform, SmartCompose

from .abstract_dataset import BaseClassificationDataset


class Standardizer(AbstractTransform):
    def __init__(self, mean: List[float], std: List[float]) -> None:
        super().__init__(order=-100)
        self.mean = mean
        self.std = std
        self._mean = torch.Tensor(mean)
        self._std = torch.Tensor(std)

    def __call__(self, x: torch.Tensor) -> torch.Tensor:
        return self.encode(x)

    def encode(self, x: torch.Tensor) -> torch.Tensor:
        return (x - self._mean) / (self._std)

    def decode(self, x: torch.Tensor) -> torch.Tensor:  # pragma: no cover
        return x * self._std + self._mean


class AIBO(BaseClassificationDataset):
    def __init__(
        self,
        path: str,
        features_subdir: str,
        seed: int,
        metrics: List[Union[str, DictConfig, Dict]],
        tracking_metric: Union[str, DictConfig, Dict],
        index_column: str,
        target_column: Union[str, List[str]],
        file_type: str,
        file_handler: Union[str, DictConfig, Dict],
        batch_size: int,
        inference_batch_size: Optional[int] = None,
        train_transform: Optional[SmartCompose] = None,
        dev_transform: Optional[SmartCompose] = None,
        test_transform: Optional[SmartCompose] = None,
        stratify: Optional[List[str]] = None,
        standardize: bool = False,
        aibo_task: str = "2cl",
    ) -> None:
        """FAU AIBO dataset.

        Args:
            path: Root path to the dataset.
            features_subdir: Subdirectory containing the features.
            seed: Seed for reproducibility.
            metrics: List of metrics to calculate.
            tracking_metric: Metric to track.
            index_column: Index column of the dataframe.
            target_column: Target column of the dataframe.
            file_type: File type of the features.
            file_handler: File handler to load the data.
            batch_size: Batch size.
            inference_batch_size: Inference batch size. If None, defaults to
                batch_size. Defaults to None.
            train_transform: Transform to apply to the training set.
                Defaults to None.
            dev_transform: Transform to apply to the development set.
                Defaults to None.
            test_transform: Transform to apply to the test set.
                Defaults to None.
            stratify: Columns to stratify the dataset on. Defaults to None.
            standardize: Whether to standardize the data. Defaults to False.
            aibo_task: Task to load in ["2cl", "5cl"]. Defaults to "2cl".
        """
        self._assert_choice(aibo_task, ["2cl", "5cl"], "aibo_task")
        self.aibo_task = aibo_task
        super().__init__(
            path=path,
            features_subdir=features_subdir,
            seed=seed,
            metrics=metrics,
            tracking_metric=tracking_metric,
            index_column=index_column,
            target_column=target_column,
            file_type=file_type,
            file_handler=file_handler,
            batch_size=batch_size,
            inference_batch_size=inference_batch_size,
            train_transform=train_transform,
            dev_transform=dev_transform,
            test_transform=test_transform,
            stratify=stratify,
        )
        self.standardize = standardize
        if self.standardize:
            train_data = torch.cat([x for x, *_ in self.train_dataset])
            print(train_data.mean(0).shape, train_data.std(0).shape)
            standardizer = Standardizer(
                mean=train_data.mean(0).tolist(),
                std=train_data.std(0).tolist(),
            )

            self.train_transform += standardizer
            self.dev_transform += standardizer
            self.test_transform += standardizer

            self.train_dataset = self._init_dataset(
                self.df_train, self.train_transform
            )

    def load_dataframes(
        self,
    ) -> Tuple[pd.DataFrame, pd.DataFrame, pd.DataFrame]:
        return (
            pd.read_csv(
                os.path.join(self.path, f"train_{self.aibo_task}.csv")
            ),
            pd.read_csv(os.path.join(self.path, f"dev_{self.aibo_task}.csv")),
            pd.read_csv(os.path.join(self.path, f"test_{self.aibo_task}.csv")),
        )

    @staticmethod
    def download(path: str) -> None:  # pragma: no cover
        """
        Download the FAU AIBO dataset.

        As the AIBO dataset is private, this method does not download the
        dataset but rather prepares the file structure expected by the
        preprocessing routines.

        In the specified path, the following directories and files are expected:

        - `default/`: Directory containing .wav files.
        - `chunk_labels_2cl_corpus.txt`: File containing the file names and
          corresponding labels for the 2-class classification task.
        - `chunk_labels_5cl_corpus.txt`: File containing the file names and
          corresponding labels for the 5-class classification task.

        Produces the following splits for both tasks (2cl and 5cl):

        - `train_{task}.csv`: Training split of all speakers of the
          Ohm-Gymnasium with the exception of the last two speakers.
        - `dev_{task}.csv`: Development split of the last two speakers of the
          Ohm-Gymnasium.
        - `test_{task}.csv`: Test split of all speakers of the
          Montessori-Schule.

        For more information on the dataset and dataset split, see:
        https://doi.org/10.1109/ICME51207.2021.9428217

        Args:
            path: Path to the directory to download the dataset to.
        """
        if os.path.isfile(os.path.join(path, "train_2cl.csv")):
            return
        if not os.path.isdir(os.path.join(path, "default")):
            raise ValueError(
                f"Directory 'default' does not exist in '{path}'."
            )
        if not os.path.isfile(
            os.path.join(path, "chunk_labels_2cl_corpus.txt")
            or os.path.isfile(
                os.path.join(path, "chunk_labels_5cl_corpus.txt")
            )
        ):
            raise ValueError(
                f"File 'chunk_labels_2cl_corpus.txt' or "
                f"'chunk_labels_5cl_corpus.txt' does not exist in '{path}'."
            )

        for task in ["2cl", "5cl"]:
            df = pd.read_csv(
                os.path.join(path, f"chunk_labels_{task}_corpus.txt"),
                header=None,
                sep=" ",
            )
            df = df.rename(columns={0: "id", 1: "class", 2: "conf"})
            df["file"] = df["id"].apply(lambda x: x + ".wav")
            df["school"] = df["id"].apply(lambda x: x.split("_")[0])
            df["speaker"] = df["id"].apply(lambda x: x.split("_")[1])
            df = df.set_index("file")
            df_test = df.loc[df["school"] == "Mont"]
            df_train_dev = df.loc[df["school"] == "Ohm"]
            speakers = sorted(df_train_dev["speaker"].unique())
            df_train = df_train_dev.loc[
                df_train_dev["speaker"].isin(speakers[:-2])
            ]
            df_dev = df_train_dev.loc[
                df_train_dev["speaker"].isin(speakers[-2:])
            ]

            df_train.to_csv(os.path.join(path, f"train_{task}.csv"))
            df_dev.to_csv(os.path.join(path, f"dev_{task}.csv"))
            df_test.to_csv(os.path.join(path, f"test_{task}.csv"))



from typing import Optional

import torch
import torch.nn.functional as F

from .abstract_model import AbstractModel
from .utils import ConvBlock, init_bn, init_layer, load_transfer_weights


class Cnn10(AbstractModel):
    def __init__(
        self,
        output_dim: int,
        sigmoid_output: bool = False,
        sigmoid_predictions: bool = False,
        segmentwise: bool = False,
        in_channels: int = 1,
        transfer: Optional[str] = None,
    ) -> None:
        """CNN10 PANN model. For more information see:
        https://doi.org/10.48550/arXiv.1912.10211

        Args:
            output_dim: Output dimension of the model.
            sigmoid_output: Whether to apply sigmoid activation to the output.
                Defaults to False.
            sigmoid_predictions: Whether to apply sigmoid activation during
                inference. Defaults to False.
            segmentwise: Whether to use segmentwise path or clipwise path.
                Defaults to False.
            in_channels: Number of input channels. Defaults to 1.
            transfer: Link to the weights to transfer. If None, the weights
                weights will be randomly initialized. Defaults to None.
        """
        super().__init__(output_dim)
        self.sigmoid_output = sigmoid_output
        self.sigmoid_predictions = sigmoid_predictions
        self.segmentwise = segmentwise
        self.in_channels = in_channels
        self.transfer = transfer
        self.bn0 = torch.nn.BatchNorm2d(64)

        self.conv_block1 = ConvBlock(in_channels=in_channels, out_channels=64)
        self.conv_block2 = ConvBlock(in_channels=64, out_channels=128)
        self.conv_block3 = ConvBlock(in_channels=128, out_channels=256)
        self.conv_block4 = ConvBlock(in_channels=256, out_channels=512)

        self.fc1 = torch.nn.Linear(512, 512, bias=True)
        self.out = torch.nn.Linear(512, output_dim, bias=True)

        self.init_weight()
        if self.transfer is not None:
            load_transfer_weights(self, self.transfer)

    def init_weight(self) -> None:
        init_bn(self.bn0)
        init_layer(self.fc1)
        init_layer(self.out)

    def get_embedding(self, x: torch.Tensor) -> torch.Tensor:
        x = x.transpose(1, 3)
        x = self.bn0(x)
        x = x.transpose(1, 3)

        x = self.conv_block1(x, pool_size=(2, 2), pool_type="avg")
        x = F.dropout(x, p=0.2, training=self.training)
        x = self.conv_block2(x, pool_size=(2, 2), pool_type="avg")
        x = F.dropout(x, p=0.2, training=self.training)
        x = self.conv_block3(x, pool_size=(2, 2), pool_type="avg")
        x = F.dropout(x, p=0.2, training=self.training)
        x = self.conv_block4(x, pool_size=(2, 2), pool_type="avg")
        x = F.dropout(x, p=0.2, training=self.training)
        x = torch.mean(x, dim=3)

        if self.segmentwise:
            return self.segmentwise_path(x)
        else:
            return self.clipwise_path(x)

    def segmentwise_path(self, x: torch.Tensor) -> torch.Tensor:
        x1 = F.max_pool1d(x, kernel_size=3, stride=1, padding=1)
        x2 = F.avg_pool1d(x, kernel_size=3, stride=1, padding=1)
        x = x1 + x2
        x = F.dropout(x, p=0.5, training=self.training)
        x = x.transpose(1, 2)
        x = F.relu_(self.fc1(x))
        x = F.dropout(x, p=0.5, training=self.training)
        return x

    def clipwise_path(self, x: torch.Tensor) -> torch.Tensor:
        (x1, _) = torch.max(x, dim=2)
        x2 = torch.mean(x, dim=2)
        x = x1 + x2

        x = F.relu_(self.fc1(x))
        return x

    def embeddings(self, x: torch.Tensor) -> torch.Tensor:
        return self.get_embedding(x)

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        x = self.embeddings(x)
        x = self.out(x)
        if self.sigmoid_output:
            x = torch.sigmoid(x)
        if (not self.training) and self.sigmoid_predictions:
            # sigmoid output for multi-label classification
            if not self.sigmoid_output:  # avoid double op
                x = torch.sigmoid(x)
        return x



import argparse
from typing import List, Optional

import autrainer
from autrainer.core.scripts import (
    AbstractScript,
    CommandLineError,
    CreateScript,
    DeleteFailedScript,
    DeleteStatesScript,
    FetchScript,
    GroupScript,
    InferenceScript,
    ListScript,
    PostprocessScript,
    PreprocessScript,
    ShowScript,
    TrainScript,
    print_help_on_error,
)


class CLI:
    def __init__(self) -> None:
        self.scripts = self._init_scripts()
        self.parser = self._init_parser()

    def _init_scripts(self) -> List[AbstractScript]:
        return [
            # configuration management
            CreateScript(),
            ListScript(),
            ShowScript(),
            # preprocessing
            FetchScript(),
            PreprocessScript(),
            # training
            TrainScript(),
            # inference
            InferenceScript(),
            # postprocessing
            PostprocessScript(),
            DeleteFailedScript(),
            DeleteStatesScript(),
            GroupScript(),
        ]

    def _init_parser(self) -> argparse.ArgumentParser:
        parser = argparse.ArgumentParser(
            prog="autrainer",
            description=(
                "A Modular and Extensible Deep Learning Toolkit "
                "for Computer Audition Tasks."
            ),
        )
        parser.error = print_help_on_error(parser)
        parser.add_argument(
            "-v",
            "--version",
            action="version",
            version=f"%(prog)s {autrainer.__version__}",
        )
        subparsers = parser.add_subparsers(
            dest="command",
            metavar="<command>",
        )
        for c in self.scripts:
            p = subparsers.add_parser(
                c.command,
                help=c.description,
            )
            p = CLI._setup_subparser(p, c)
        return parser

    @staticmethod
    def _setup_subparser(
        parser: argparse.ArgumentParser,
        command: AbstractScript,
    ) -> None:
        if command.extended_description:
            parser.description = (
                command.description + "\n" + command.extended_description
            )
        else:
            parser.description = command.description
        parser.epilog = command.epilog
        parser.formatter_class = argparse.RawTextHelpFormatter
        parser.error = print_help_on_error(parser)
        command.init_parser(parser)
        command.add_arguments()

    def _get_command(self, name: str) -> Optional[AbstractScript]:
        return next((c for c in self.scripts if c.command == name), None)

    def main(self) -> None:
        args, unknown = self.parser.parse_known_args()
        args = vars(args)
        command = self._get_command(args.pop("command", None))
        if command:
            command.run(args, unknown)
        else:
            raise CommandLineError(self.parser, "No command provided.")


def main() -> None:
    cli = CLI()
    try:
        cli.main()
    except CommandLineError as e:
        e.handle()


def get_parser() -> argparse.ArgumentParser:
    cli = CLI()  # pragma: no cover
    return cli.parser  # pragma: no cover


if __name__ == "__main__":
    main()  # pragma: no cover



import glob
import os
import sys
from typing import Any, Callable, Dict, List, Optional, Tuple, Union
import warnings

import audobject
from omegaconf import OmegaConf
import pandas as pd
import torch
import yaml

import autrainer
from autrainer.core.utils import set_device
from autrainer.datasets.utils import (
    AbstractFileHandler,
    AbstractTargetTransform,
)
from autrainer.models import AbstractModel
from autrainer.transforms import SmartCompose


class Inference:
    def __init__(
        self,
        model_path: str,
        checkpoint: str = "_best",
        device: str = "cpu",
        preprocess_cfg: Optional[str] = "default",
        window_length: Optional[float] = None,
        stride_length: Optional[float] = None,
        min_length: Optional[float] = None,
        sample_rate: Optional[int] = None,
    ) -> None:
        """Inference class for audio models.

        Args:
            model_path: Local model directory containing the model.yaml,
                file_handler.yaml, target_transform.yaml, and
                inference_transform.yaml files.
            checkpoint: Checkpoint directory containing a model.pt file.
                Defaults to "_best".
            device: Device to run inference on. Defaults to "cpu".
            preprocess_cfg: Preprocessing configuration file. If "default",
                the default preprocessing pipeline used during training is
                applied. If None, no preprocessing is applied. Defaults to
                "default".
            window_length: Window length in seconds for sliding window
                inference. Defaults to None.
            stride_length: Stride length in seconds for sliding window
                inference. Defaults to None.
            min_length: Minimum length of an audio file in seconds.
                Audio files shorter than this will be padded with zeros.
                Defaults to None.
            sample_rate: Sample rate of the audio files in Hz.
                Defaults to None.
        """
        self._model_path = model_path
        self._checkpoint = checkpoint
        self._device = set_device(device)
        self._preprocess_cfg = preprocess_cfg
        self._window_length = window_length
        self._stride_length = stride_length
        self._min_length = min_length
        self._sample_rate = sample_rate
        sys.path.append(os.getcwd())

        self.model: AbstractModel = audobject.from_yaml(
            os.path.join(self._model_path, "model.yaml")
        )
        with warnings.catch_warnings():
            warnings.simplefilter("ignore", UserWarning)
            state = torch.load(
                os.path.join(self._model_path, self._checkpoint, "model.pt"),
                map_location="cpu",
                weights_only=True,
            )
        self.model.load_state_dict(state)
        self.model.to(self._device)
        self.model.eval()
        self.file_handler: AbstractFileHandler = audobject.from_yaml(
            os.path.join(self._model_path, "file_handler.yaml")
        )
        self.target_transform: AbstractTargetTransform = audobject.from_yaml(
            os.path.join(self._model_path, "target_transform.yaml")
        )
        self.inference_transform: SmartCompose = audobject.from_yaml(
            os.path.join(self._model_path, "inference_transform.yaml")
        )

        self.preprocess_pipeline = SmartCompose([])
        if self._preprocess_cfg == "default":
            self.file_handler: AbstractFileHandler = audobject.from_yaml(
                os.path.join(self._model_path, "preprocess_file_handler.yaml")
            )
            self.preprocess_pipeline: SmartCompose = audobject.from_yaml(
                os.path.join(self._model_path, "preprocess_pipeline.yaml")
            )
        elif self._preprocess_cfg is not None:
            preprocess_cfg = OmegaConf.to_container(
                OmegaConf.load(self._preprocess_cfg)
            )
            self.file_handler = autrainer.instantiate_shorthand(
                preprocess_cfg["file_handler"]
            )
            pipeline = [
                autrainer.instantiate_shorthand(p)
                for p in preprocess_cfg["pipeline"]
            ]
            self.preprocess_pipeline = SmartCompose(pipeline)

    def predict_directory(
        self,
        directory: str,
        extension: str,
        recursive: bool = False,
    ) -> pd.DataFrame:
        """Obtain the model predictions for all files in a directory.

        Args:
            directory: Path to the directory containing audio files.
            extension: File extension of the audio files.
            recursive: Whether to search recursively for audio files in
                subdirectories. Defaults to False.

        Returns:
            DataFrame containing the filename, prediction, and output for
            each file. If sliding window inference is used, the offset is
            additionally included.
        """
        files = self._collect_files(directory, extension, recursive)
        records = []
        for file in files:
            prediction, output = self.predict_file(file)
            if isinstance(prediction, dict):
                for (offset, pred), out in zip(
                    prediction.items(), output.values()
                ):
                    records.append(
                        {
                            "filename": os.path.relpath(file, directory),
                            "offset": offset,
                            "prediction": pred,
                            "output": out.squeeze().tolist(),
                        }
                    )
            else:
                records.append(
                    {
                        "filename": os.path.relpath(file, directory),
                        "prediction": prediction,
                        "output": output.squeeze().tolist(),
                    }
                )
        return pd.DataFrame(records)

    def embed_directory(
        self,
        directory: str,
        extension: str,
        recursive: bool = False,
    ) -> pd.DataFrame:
        """Obtain the model embeddings for all files in a directory.

        Args:
            directory: Path to the directory containing audio files.
            extension: File extension of the audio files.
            recursive: Whether to search recursively for audio files in
                subdirectories. Defaults to False.

        Returns:
            DataFrame containing the filename and embedding for each file.
            If sliding window inference is used, the offset is additionally
            included.
        """
        files = self._collect_files(directory, extension, recursive)
        records = []
        for file in files:
            embedding = self.embed_file(file)
            if isinstance(embedding, dict):
                for offset, emb in embedding.items():
                    records.append(
                        {
                            "filename": os.path.relpath(file, directory),
                            "offset": offset,
                            "embedding": emb,
                        }
                    )
            else:
                records.append(
                    {
                        "filename": os.path.relpath(file, directory),
                        "embedding": embedding,
                    }
                )
        return pd.DataFrame(records)

    def predict_file(
        self,
        file: str,
    ) -> Union[
        Tuple[Any, torch.Tensor],
        Tuple[Dict[str, Any], Dict[str, torch.Tensor]],
    ]:
        """Obtain the model prediction for a single file.

        Args:
            file: Path to the audio file.

        Returns:
            Model prediction and output for the file. If sliding window
            inference is used, the prediction is a dictionary with the
            offset as the key.
        """
        return self._delegate_file(file, self._predict, self._predict_windowed)

    def embed_file(
        self,
        file: str,
    ) -> Union[torch.Tensor, Dict[str, torch.Tensor]]:
        """Obtain the model embedding for a single file.

        Args:
            file: Path to the audio file.

        Returns:
            Model embedding for the file. If sliding window inference is used,
            the embedding is a dictionary with the offset as the key.
        """
        return self._delegate_file(file, self._embed, self._embed_windowed)

    @staticmethod
    def save_prediction_yaml(results: pd.DataFrame, output_dir: str) -> None:
        """Save the prediction results to a YAML file.

        Creates a human-readable YAML file with the prediction results.

        Args:
            results: DataFrame containing the results.
            output_dir: Output directory to save the results to.
        """
        os.makedirs(output_dir, exist_ok=True)
        if "offset" in results.columns:
            results = (
                results.set_index(["filename", "offset"])["prediction"]
                .unstack(0)
                .to_dict()
            )
        else:
            results = results.set_index("filename")["prediction"].to_dict()
        with open(os.path.join(output_dir, "results.yaml"), "w") as fp:
            yaml.dump(results, fp)

    @staticmethod
    def save_prediction_results(
        results: pd.DataFrame,
        output_dir: str,
    ) -> None:
        """Save the prediction results to a CSV file.

        Creates a CSV file with the model predictions and outputs.

        Args:
            results: DataFrame containing the results.
            output_dir: Output directory to save the results to.
        """
        os.makedirs(output_dir, exist_ok=True)
        results.to_csv(os.path.join(output_dir, "results.csv"), index=False)

    @staticmethod
    def save_embeddings(
        results: pd.DataFrame,
        output_dir: str,
        input_extension: str,
    ) -> None:
        """Save the embeddings as torch tensors.

        Saves the embeddings to the output directory with the same filename
        as each audio file.

        Args:
            results: DataFrame containing the embeddings.
            output_dir: Output directory to save the embeddings to.
            input_extension: File extension of the input audio files to replace
                with ".pt".
        """
        os.makedirs(output_dir, exist_ok=True)

        def _create_output_name(row, offset: bool = False) -> str:
            n = row["filename"].replace("." + input_extension, "")
            n += f"-{row['offset']}" if offset else ""
            return n + ".pt"

        for _, row in results.iterrows():
            output_path = os.path.join(
                output_dir,
                _create_output_name(row, "offset" in results.columns),
            )
            torch.save(row["embedding"], output_path)

    def _delegate_file(
        self,
        file: str,
        fn: Callable,
        window_fn: Callable,
    ) -> Union[Tuple[Any, torch.Tensor], torch.Tensor]:
        x = self.file_handler.load(file)
        if self._window_length and self._stride_length and self._sample_rate:
            return window_fn(x)
        else:
            return fn(x)

    def _collect_files(self, directory: str, extension: str, recursive: bool):
        pattern = f"**/*.{extension}" if recursive else f"*.{extension}"
        return glob.glob(os.path.join(directory, pattern), recursive=True)

    def _preprocess_file(self, x: Union[torch.Tensor]) -> torch.Tensor:
        x = self._pad_audio(x)
        x = self.preprocess_pipeline(x, 0)
        x = self.inference_transform(x, 0)
        x = x.unsqueeze(0).to(self._device)
        return x

    def _predict(self, x: Union[torch.Tensor]) -> Tuple[Any, torch.Tensor]:
        x = self._preprocess_file(x)
        with torch.inference_mode():
            output = self.model(x).cpu()
        prediction = self.target_transform.predict_batch(output)
        prediction = self.target_transform.decode(prediction)
        return prediction, output

    def _embed(self, x: Union[torch.Tensor]) -> torch.Tensor:
        x = self._preprocess_file(x)
        with torch.inference_mode():
            embedding = self.model.embeddings(x).cpu().squeeze()
        return embedding

    def _create_windows(self, x: torch.Tensor) -> Tuple[int, int, List[int]]:
        w_len = int(self._window_length * self._sample_rate)
        s_len = int(self._stride_length * self._sample_rate)
        num_windows = (x.shape[1] - w_len) // s_len + 1
        return w_len, s_len, num_windows

    def _predict_windowed(
        self,
        x: Union[torch.Tensor],
    ) -> Tuple[Dict[str, Any], Dict[str, torch.Tensor]]:
        results = {}
        outputs = {}
        w_len, s_len, num_windows = self._create_windows(x)
        for i in range(num_windows):
            start_idx = i * s_len
            end_idx = min(start_idx + w_len, x.shape[1])
            prediction, output = self._predict(x[:, start_idx:end_idx])
            start_time = start_idx / self._sample_rate
            end_time = end_idx / self._sample_rate
            results[f"{start_time:.2f}-{end_time:.2f}"] = prediction
            outputs[f"{start_time:.2f}-{end_time:.2f}"] = output
        majority = self.target_transform.majority_vote(list(results.values()))
        results["majority"] = majority
        outputs["majority"] = torch.empty(0)
        return results, outputs

    def _embed_windowed(
        self,
        x: Union[torch.Tensor],
    ) -> Dict[str, torch.Tensor]:
        results = {}
        w_len, s_len, num_windows = self._create_windows(x)
        for i in range(num_windows):
            start_idx = i * s_len
            end_idx = min(start_idx + w_len, x.shape[1])
            embedding = self._embed(x[:, start_idx:end_idx])
            start_time = start_idx / self._sample_rate
            end_time = end_idx / self._sample_rate
            results[f"{start_time:.2f}-{end_time:.2f}"] = embedding
        return results

    def _pad_audio(self, x: Union[torch.Tensor]) -> Union[torch.Tensor]:
        if not self._min_length:
            return x
        min_length = int(self._min_length * self._sample_rate)
        if x.shape[1] >= min_length:
            return x
        return torch.nn.functional.pad(
            x,
            (0, min_length - x.shape[1]),
            value=0,
        )

    def __repr__(self) -> str:
        s = (
            f"File handler:\n{self.file_handler}"
            f"Data transform:\n{self.inference_transform}"
            f"Model:\n{self.model}"
            f"Label transform:\n{self.target_transform}"
        )
        return s



from functools import wraps
import sys
from typing import Any, Callable, Optional, TypeVar

from .command_line_error import CommandLineError


F = TypeVar("F", bound=Callable[..., Any])
TQDM_LINE_ENDINGS = ["it/s]", "s/it]", "B/s]"]


def run_hydra_cmd(
    cmd: str,
    override_kwargs: Optional[dict] = None,
    config_name: str = "config",
    config_path: Optional[str] = None,
    cmd_prefix: str = "autrainer",
) -> None:
    """Run a Hydra command in a subprocess and print the output.

    Args:
        cmd: The Hydra command to run.
        override_kwargs: Additional Hydra override arguments to pass to the
            train script.
        config_name: The name of the config (usually the file name without the
            .yaml extension). Defaults to "config".
        config_path: The config path, a directory where Hydra will search for
            config files. If config_path is None no directory is added to the
            search path. Defaults to None.
        cmd_prefix: The command prefix to use. Defaults to "autrainer".
    """
    import subprocess

    cmd = f"{cmd_prefix} {cmd} -cn {config_name}.yaml"
    if config_path is not None:
        cmd += f" -cp {config_path}"
    if override_kwargs is not None:
        for key, value in override_kwargs.items():
            if isinstance(value, list):
                value = ",".join(map(str, value))
            if not isinstance(value, (str, int, float, bool)):
                raise CommandLineError(
                    f"Hydra override argument values must be of type list, "
                    f"str, int, float, or bool, but got {type(value)}."
                )
            cmd += f" {key}={value}"

    with subprocess.Popen(
        cmd,
        stdout=subprocess.PIPE,
        stderr=subprocess.STDOUT,
        shell=True,
        bufsize=1,
        encoding="utf-8",
        errors="replace",
        universal_newlines=False,
    ) as process:
        tqdm_line = False
        for line in process.stdout:
            line = line.rstrip()

            # TODO: find a better way to handle tqdm output
            if any(line.endswith(ending) for ending in TQDM_LINE_ENDINGS):
                print("\r" + line, end="")
                tqdm_line = True
            elif tqdm_line:
                print("\n" + line)
                tqdm_line = False
            else:
                print(line)

    process.stdout.close()
    process.wait()


def running_in_notebook() -> bool:
    return any("jupyter" in arg or "ipykernel" in arg for arg in sys.argv)


def catch_cli_errors(func: F) -> F:
    @wraps(func)
    def wrapper(*args: Any, **kwargs: Any) -> Any:
        try:
            return func(*args, **kwargs)
        except CommandLineError as e:
            if not running_in_notebook():
                raise e
            print(e.message, file=sys.stderr)

    return wrapper


def add_hydra_args_to_sys(
    override_kwargs: Optional[dict] = None,
    config_name: str = "config",
    config_path: Optional[str] = None,
) -> None:
    if config_name.replace(".yaml", "") != "config":
        sys.argv.extend(["-cn", config_name])
    if config_path is not None:
        sys.argv.extend(["-cp", config_path])
    if override_kwargs is not None:
        sys.argv.extend(f"{k}={v}" for k, v in override_kwargs.items())



from abc import ABC, abstractmethod
import os
from pathlib import Path
import re
from typing import List, Optional, Tuple


MODEL_FILES = [
    "model.yaml",
    "file_handler.yaml",
    "target_transform.yaml",
    "inference_transform.yaml",
    "preprocess_file_handler.yaml",
    "preprocess_pipeline.yaml",
]
SAVE_FILES = tuple(["model.pt"] + MODEL_FILES)


class AbstractModelPath(ABC):
    def __init__(self, model_path: str) -> None:
        """Abstract class for model path construction.

        Args:
            model_path: Model path.
        """
        self.model_path = model_path

    @abstractmethod
    def verify(self) -> None:
        """Verify the model path.

        Raises:
            ValueError: If the model path is invalid.
        """

    @abstractmethod
    def create_model_path(self) -> str:
        """Create the directory of the model.

        Returns:
            The directory of the model.
        """

    def get_model_path(self) -> str:
        """Get the directory of the model.

        Returns:
            The directory of the model.
        """
        path = self.create_model_path()
        for file in MODEL_FILES:
            if not os.path.isfile(os.path.join(path, file)):
                raise ValueError(
                    f"Invalid model directory: '{file}' not found in '{path}'"
                )
        return path


class LocalModelPath(AbstractModelPath):
    def __init__(self, model_path: str) -> None:
        super().__init__(model_path)
        self.verify()

    def verify(self) -> None:
        if not os.path.isdir(self.model_path):
            raise ValueError(
                f"Invalid local model directory: '{self.model_path}'"
            )

    def create_model_path(self) -> str:
        return self.model_path


class HubModelPath(AbstractModelPath):
    def __init__(self, model_path: str) -> None:
        super().__init__(model_path)
        self.verify()

    def verify(self) -> None:
        """Verify the given hugging face model path.

        Verification includes checking the validity of the repository, revision,
        and repository subdirectory in the model path.
        The model files are also verified to ensure that a single model
        directory is found.

        Raises:
            ValueError: If the model path is invalid.
            ValueError: If the repository, revision or repository subdirectory
            is not found.
            ValueError: If the model directory is not found or multiple model
            directories are found.
        """
        import torch

        repo_id, revision, subdir, local_dir = self._parse_hub_path()
        files = self._verify_hub_path(repo_id, revision, subdir)

        model_files = [f for f in files if os.path.basename(f) == "model.yaml"]

        if subdir:
            model_files = [f for f in model_files if f.startswith(subdir)]

        self._verify_model_files(repo_id, revision, subdir, model_files)

        model_file_loc = os.path.dirname(model_files[0])

        self.repo_id = repo_id
        self.revision = revision
        self.local_dir = Path(
            local_dir
            or os.path.join(
                torch.hub.get_dir(),
                "autrainer",
                repo_id.replace("/", "--") + f"--{revision or 'main'}",
            )
        ).as_posix()
        self.files = [
            Path(f).relative_to(model_file_loc).as_posix()
            for f in files
            if f.startswith(model_file_loc) and f.endswith(SAVE_FILES)
        ]
        self.subdir = model_file_loc

    def create_model_path(self) -> str:
        """Create the local directory of the model.

        The model is downloaded to the local directory if it is not already
        present.
        If no local directory is specified, the model is downloaded to the
        default torch.hub directory.

        Returns:
            Local directory of the model.
        """
        self._download()
        return os.path.join(self.local_dir, self.subdir)

    def _parse_hub_path(
        self,
    ) -> Tuple[str, Optional[str], Optional[str], Optional[str]]:
        # pattern: hf:repo_id[@revision][:subdir]#local_dir
        pattern = r"^hf:([^@:#]+)(?:@([^:#]+))?(?::([^#]+))?(?:#(.+))?$"

        match = re.match(pattern, self.model_path)
        if not match:
            raise ValueError(
                f"Invalid hugging face path format: '{self.model_path}'"
            )

        return match.groups()

    @staticmethod
    def _verify_hub_path(
        repo_id: str,
        revision: Optional[str],
        subdir: Optional[str],
    ) -> List[str]:
        from huggingface_hub import HfApi

        try:
            hf_api = HfApi(endpoint=os.environ.get("HF_ENDPOINT"))
            files = hf_api.list_repo_files(
                repo_id,
                revision=revision,
            )
        except Exception:
            raise ValueError(
                f"Invalid hugging face repo id: '{repo_id}'"
                + (f" or revision: '{revision}'" if revision else "")
            )
        if subdir is not None and f"{subdir}/model.yaml" not in files:
            raise ValueError(
                f"Subdirectory '{subdir}' not found in repo '{repo_id}' "
                "or is not a valid model directory"
            )
        return files

    @staticmethod
    def _verify_model_files(
        repo_id: str,
        revision: Optional[str],
        subdir: Optional[str],
        model_files: List[str],
    ) -> None:
        if not model_files:
            raise ValueError(
                f"No model directory found in repo '{repo_id}'"
                + (f" at revision '{revision}'" if revision else "")
                + (f" with subdir '{subdir}'" if subdir else "")
            )
        if len(model_files) > 1:
            raise ValueError(
                f"Multiple model directories found in repo '{repo_id}'"
            )

    def _download(self) -> None:
        from huggingface_hub import HfApi
        from tqdm import tqdm

        from autrainer.core.utils import silence

        os.makedirs(self.local_dir, exist_ok=True)
        self.files = [
            f
            for f in self.files
            if not os.path.isfile(os.path.join(self.local_dir, self.subdir, f))
        ]
        if not self.files:
            return
        hf_api = HfApi(endpoint=os.environ.get("HF_ENDPOINT"))
        pbar = tqdm(self.files, desc="Downloading:")
        for file in pbar:
            pbar.set_description(f"Downloading: {file}")
            with silence():
                hf_api.hf_hub_download(
                    repo_id=self.repo_id,
                    filename=file,
                    subfolder=self.subdir,
                    revision=self.revision,
                    local_dir=self.local_dir,
                )


def get_model_path(model_path: str) -> str:
    if model_path.startswith("hf:"):
        return HubModelPath(model_path).get_model_path()
    return LocalModelPath(model_path).get_model_path()



import warnings

import torch
from transformers import Wav2Vec2Model

from .abstract_model import AbstractModel
from .ffnn import FFNN


class W2V2Backbone(AbstractModel):
    def __init__(
        self,
        model_name,
        freeze_extractor: bool = True,
        time_pooling: bool = True,
    ) -> None:
        self.model_name = model_name
        self.freeze_extractor = freeze_extractor
        self.time_pooling = time_pooling
        with warnings.catch_warnings():
            warnings.simplefilter("ignore")
            model = Wav2Vec2Model.from_pretrained(self.model_name)
        super().__init__(output_dim=model.config.hidden_size)
        self.model = model
        if self.freeze_extractor:
            self.model.freeze_feature_encoder()

    def embeddings(self, x: torch.Tensor) -> torch.Tensor:
        x = self.model(x)["last_hidden_state"]
        if self.time_pooling:
            x = x.mean(1)
        return x

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        return self.embeddings(x)


class W2V2FFNN(AbstractModel):
    def __init__(
        self,
        output_dim: int,
        model_name: str,
        freeze_extractor: bool,
        hidden_size: int,
        num_layers: int = 2,
        sigmoid: bool = False,
        softmax: bool = False,
        dropout: float = 0.5,
    ) -> None:
        """Wav2Vec2 model with FFNN frontend adapted for audio classification.
        For more information, see:
        https://huggingface.co/docs/transformers/model_doc/wav2vec2

        Args:
            output_dim: Output dimension of the FFNN.
            model_name: Name of the model loaded from Huggingface.
            freeze_extractor: Whether to freeze the feature extractor.
            hidden_size: Hidden size of the FFNN.
            num_layers: Number of layers of the FFNN. Defaults to 2.
            sigmoid: Whether to apply sigmoid activation. Defaults to False.
            softmax: Whether to apply softmax activation. Defaults to False.
            dropout: Dropout rate. Defaults to 0.5.
        """
        super().__init__(output_dim)
        self.model_name = model_name
        self.freeze_extractor = freeze_extractor
        self.hidden_size = hidden_size
        self.num_layers = num_layers
        self.sigmoid = sigmoid
        self.softmax = softmax
        self.dropout = dropout
        self.backbone = W2V2Backbone(
            model_name=model_name,
            freeze_extractor=freeze_extractor,
            time_pooling=True,
        )
        self.frontend = FFNN(
            input_size=self.backbone.output_dim,
            hidden_size=hidden_size,
            output_dim=output_dim,
            num_layers=num_layers,
            sigmoid=sigmoid,
            softmax=softmax,
            dropout=dropout,
        )

    def embeddings(self, x: torch.Tensor) -> torch.Tensor:
        return self.backbone(x.squeeze(1))

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        return self.frontend(self.embeddings(x))



from .model_paths import (
    AbstractModelPath,
    HubModelPath,
    LocalModelPath,
    get_model_path,
)
from .serving import Inference


__all__ = [
    "AbstractModelPath",
    "HubModelPath",
    "Inference",
    "LocalModelPath",
    "get_model_path",
]



from .abstract_logger import (
    EXPORT_ARTIFACTS,
    AbstractLogger,
    get_params_to_export,
)
from .fallback_logger import FallbackLogger
from .mlflow_logger import MLFlowLogger
from .tensorboard_logger import TensorBoardLogger


__all__ = [
    "EXPORT_ARTIFACTS",
    "AbstractLogger",
    "FallbackLogger",
    "MLFlowLogger",
    "TensorBoardLogger",
    "get_params_to_export",
]



import os
from typing import Dict, List, Union

from omegaconf import DictConfig

from autrainer.metrics import AbstractMetric

from .abstract_logger import (
    EXPORT_ARTIFACTS,
    AbstractLogger,
    get_params_to_export,
)
from .fallback_logger import FallbackLogger


try:
    from torch.utils.tensorboard import SummaryWriter

    TENSORBOARD_AVAILABLE = True
except ImportError:  # pragma: no cover
    TENSORBOARD_AVAILABLE = False  # pragma: no cover


class TensorBoardLogger(AbstractLogger):
    def __init__(
        self,
        exp_name: str,
        run_name: str,
        metrics: List[AbstractMetric],
        tracking_metric: AbstractMetric,
        artifacts: List[Union[str, Dict[str, str]]] = EXPORT_ARTIFACTS,
        output_dir: str = "runs",
    ) -> None:
        super().__init__(
            exp_name, run_name, metrics, tracking_metric, artifacts
        )
        self.output_dir = os.path.join(output_dir, exp_name, run_name)

    def setup(self) -> None:
        self.writer = SummaryWriter(log_dir=self.output_dir)

    def log_params(self, params: Union[dict, DictConfig]) -> None:
        for key, value in get_params_to_export(params).items():
            self.writer.add_text(key, str(value))
        self.writer.flush()

    def log_metrics(
        self,
        metrics: Dict[str, Union[int, float]],
        iteration=None,
    ) -> None:
        for key, value in metrics.items():
            self.writer.add_scalar(key, value, iteration)
        self.writer.flush()

    def log_timers(self, timers: Dict[str, float]) -> None:
        for key, value in timers.items():
            self.writer.add_text(key, str(value))
        self.writer.flush()

    def log_artifact(self, filename: str, path: str = "") -> None:
        # TensorBoard does not support logging artifacts
        pass

    def end_run(self) -> None:
        self.writer.close()


TensorBoardLogger = (
    TensorBoardLogger
    if TENSORBOARD_AVAILABLE
    else lambda *args, **kwargs: FallbackLogger(
        "TensorBoardLogger",
        "tensorboard",
    )
)



from typing import Dict, List, Optional

import audobject
import torch

import autrainer

from .abstract_augmentation import AbstractAugmentation


class Choice(AbstractAugmentation, audobject.Object):
    def __init__(
        self,
        choices: List[Dict],
        weights: Optional[List[float]] = None,
        order: int = 0,
        p: float = 1.0,
        generator_seed: Optional[int] = None,
    ) -> None:
        """Choose one augmentation from a list of augmentations with a given
        probability.

        The order of the augmentations in the list is not considered and is
        placed with respect to the order of the choice augmentation itself.

        Augmentations in the list must not have a collate function.

        Args:
            choices: A list of (shorthand syntax) dictionaries defining the
                augmentations to choose from.
            weights: A list of weights for each choice. If None, all
                augmentations are assigned equal weights. Defaults to None.
            order: The order of the augmentation in the transformation pipeline.
                Defaults to 0.
            p: The probability of applying the augmentation. Defaults to 1.0.
            generator_seed: The initial seed for the internal random number
                generator drawing the probability. If None, the generator is
                not seeded. Defaults to None.

        Raises:
            ValueError: If choices and weights have different lengths.
            ValueError: If any augmentation has a collate function.
        """
        super().__init__(order, p, generator_seed)
        weights = weights or [1.0] * len(choices)
        if len(choices) != len(weights):
            raise ValueError("Choices and weights must have the same length.")

        self.weights = [x / sum(weights) for x in weights]
        self.choices = choices
        self.augmentation_choices = []

        for aug in self.choices:
            if isinstance(aug, str):
                self.augmentation_choices.append(
                    {aug: {"generator_seed": self.generator_seed}}
                )
            else:
                aug_name = next(iter(aug.keys()))
                if aug[aug_name].get("generator_seed") is None:
                    aug[aug_name]["generator_seed"] = self.generator_seed
                self.augmentation_choices.append(aug)

        self.augmentation_choices: List[AbstractAugmentation] = [
            autrainer.instantiate_shorthand(
                choice,
                instance_of=AbstractAugmentation,
            )
            for choice in self.choices
        ]
        for aug in self.augmentation_choices:
            if hasattr(aug, "get_collate_fn"):
                raise ValueError(
                    "Choice augmentations must not have a collate function."
                )

    def apply(self, x: torch.Tensor, index: int = None) -> torch.Tensor:
        """Choose one augmentation from the list of augmentations based on the
        given weights.

        Args:
            x: The input tensor.
            index: The index of the input tensor in the dataset.
                Defaults to None.

        Returns:
            The augmented tensor.
        """
        weights = torch.Tensor(self.weights)
        choice = torch.multinomial(weights, 1, generator=self.g).item()
        return self.augmentation_choices[choice](x, index)



import os
import warnings

from speechbrain.inference.classifiers import EncoderClassifier
import torch

from .abstract_model import AbstractModel
from .ffnn import FFNN


class TDNNFFNN(AbstractModel):
    def __init__(
        self,
        output_dim: int,
        hidden_size: int,
        num_layers: int = 2,
        sigmoid: bool = False,
        softmax: bool = False,
        dropout: float = 0.5,
    ) -> None:
        """Time Delay Neural Network with FFNN frontend.

        Args:
            output_dim: Output dimension.
            hidden_size: Hidden size.
            num_layers: Number of layers. Defaults to 2.
            sigmoid: Whether to use sigmoid activation. Defaults to False.
            softmax: Whether to use softmax activation. Defaults to False.
            dropout: Dropout rate. Defaults to 0.5.
        """
        super().__init__(output_dim)
        self.hidden_size = hidden_size
        self.num_layers = num_layers
        self.sigmoid = sigmoid
        self.softmax = softmax
        self.dropout = dropout
        checkpoint_dir = os.path.join(torch.hub.get_dir(), "speechbrain")
        os.makedirs(checkpoint_dir, exist_ok=True)
        with warnings.catch_warnings():
            warnings.simplefilter("ignore", category=FutureWarning)
            tdnn = EncoderClassifier.from_hparams(
                source="speechbrain/spkrec-ecapa-voxceleb",
                freeze_params=False,
                savedir=checkpoint_dir,
            )
        self.backbone = tdnn.mods["embedding_model"]
        self.features = tdnn.mods["compute_features"]
        self.frontend = FFNN(
            input_size=192,  # TODO: get from model?
            hidden_size=hidden_size,
            output_dim=output_dim,
            num_layers=num_layers,
            sigmoid=sigmoid,
            softmax=softmax,
            dropout=dropout,
        )

    def embeddings(self, x: torch.Tensor) -> torch.Tensor:
        _device = x.device
        feats = self.features(x.squeeze(1).cpu())
        feats = feats.to(_device)
        embs = self.backbone(feats).squeeze(1)
        return embs

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        embs = self.embeddings(x)
        return self.frontend(embs)



from typing import Any, Dict, List, Union

import autrainer
from autrainer.transforms import SmartCompose

from .abstract_augmentation import AbstractAugmentation


class AugmentationPipeline:
    def __init__(
        self,
        pipeline: List[Union[str, Dict[str, Any]]],
        generator_seed: int = 0,
    ) -> None:
        """Initialize an augmentation pipeline.

        Args:
            pipeline: The list of augmentations to apply.
            generator_seed: Seed to pass to each augmentation for
                reproducibility if the augmentation does not have a seed.
                Defaults to 0.
        """
        self.generator_seed = generator_seed

        self.pipeline = []
        for aug in pipeline:
            if isinstance(aug, str):
                self.pipeline.append(
                    {aug: {"generator_seed": self.generator_seed}}
                )
            else:
                aug_name = next(iter(aug.keys()))
                if aug[aug_name].get("generator_seed") is None:
                    aug[aug_name]["generator_seed"] = self.generator_seed
                self.pipeline.append(aug)

        self.pipeline: List[AbstractAugmentation] = [
            autrainer.instantiate_shorthand(
                aug,
                instance_of=AbstractAugmentation,
            )
            for aug in self.pipeline
        ]

    def create_pipeline(self) -> SmartCompose:
        """Create the composed and ordered augmentation pipeline.

        Returns:
            Composed augmentation pipeline.
        """
        return SmartCompose(self.pipeline)



from typing import List, Optional, Union
import warnings

from audtorch import transforms as AT
from matplotlib.pyplot import get_cmap
import numpy as np
from omegaconf import OmegaConf
from PIL import Image
import torch
from torchaudio import transforms as TT
import torchlibrosa
from torchvision import transforms as T
from transformers import (
    ASTFeatureExtractor,
    AutoFeatureExtractor,
    Wav2Vec2FeatureExtractor,
    WhisperFeatureExtractor,
)

from .abstract_transform import AbstractTransform
from .utils import _to_numpy, _to_tensor


try:
    import opensmile

    OPENSMILE_AVAILABLE = True
except ImportError:  # pragma: no cover
    OPENSMILE_AVAILABLE = False  # pragma: no cover


FE_MAPPINGS = {
    "AST": {"fe": ASTFeatureExtractor, "padding": "max_length"},
    "Whisper": {"fe": WhisperFeatureExtractor, "padding": "max_length"},
    "W2V2": {"fe": Wav2Vec2FeatureExtractor, "padding": "longest"},
    None: {"fe": AutoFeatureExtractor, "padding": "max_length"},
}


class AnyToTensor(AbstractTransform):
    def __init__(self, order: int = -100) -> None:
        """Convert a numpy array, torch tensor, or a PIL image to a torch
        tensor.

        Args:
            order: The order of the transform in the pipeline. Defaults to -100.
        """
        super().__init__(order=order)
        self._convert_pil = T.ToTensor()

    def __call__(
        self,
        data: Union[torch.Tensor, np.ndarray, Image.Image],
    ) -> torch.Tensor:
        if isinstance(data, torch.Tensor):
            return data
        elif isinstance(data, np.ndarray):
            return _to_tensor(data)
        elif isinstance(data, Image.Image):
            return (self._convert_pil(data) * 255).to(torch.uint8)
        else:
            raise TypeError(
                "Input must be a 'torch.Tensor', 'np.ndarray' "
                f"or 'PIL.Image.Image' but got '{data.__class__.__name__}'"
            )


class NumpyToTensor(AbstractTransform):
    def __init__(self, order: int = -100) -> None:
        """Convert a numpy array to a torch tensor.

        Args:
            order: The order of the transform in the pipeline. Defaults to -100.
        """
        super().__init__(order=order)

    def __call__(self, data: Union[torch.Tensor, np.ndarray]) -> torch.Tensor:
        return _to_tensor(data)


class SpectToImage(AbstractTransform):
    def __init__(
        self,
        height: int,
        width: int,
        cmap: str = "magma",
        order: int = -90,
    ):
        """Convert a spectrogram to a 3-channel image based on a colormap.

        Args:
            height: The height of the image.
            width: The width of the image.
            cmap: The colormap to use. Defaults to "magma".
            order: The order of the transform in the pipeline. Defaults to -90.
        """
        super().__init__(order=order)
        self.height = height
        self.width = width
        self.cmap = cmap
        self._cmap = get_cmap(cmap)

    def __call__(self, data: torch.Tensor) -> torch.Tensor:
        if data.dtype == torch.uint8:
            data = data.float()
        data = data.numpy().squeeze()
        data -= data.min()
        data /= data.max()
        data = np.uint8(self._cmap(data)[..., :3] * 255)
        im = Image.fromarray(data)
        im = im.transpose(Image.FLIP_TOP_BOTTOM)
        im = im.resize((self.width, self.height))
        return torch.from_numpy(np.array(im)).transpose(0, 2).to(torch.uint8)


class PannMel(AbstractTransform):
    def __init__(
        self,
        window_size: int,
        hop_size: int,
        sample_rate: int,
        fmin: int,
        fmax: int,
        mel_bins: int,
        ref: float,
        amin: float,
        top_db: int,
        order: int = -90,
    ):
        """Create a log-Mel spectrogram from an audio signal analogous to
        the Pretrained Audio Neural Networks (PANN) implementation.

        For more information, see:
        https://doi.org/10.48550/arXiv.1912.10211

        Args:
            window_size: The size of the window.
            hop_size: Hop length.
            sample_rate: The sample rate of the audio signal.
            fmin: The minimum frequency.
            fmax: The maximum frequency.
            mel_bins: The number of mel bins.
            ref: The reference amplitude.
            amin: The minimum amplitude.
            top_db: The top decibel.
            order: The order of the transform in the pipeline. Defaults to -90.
        """
        self.window_size = window_size
        self.hop_size = hop_size
        self.sample_rate = sample_rate
        self.fmin = fmin
        self.fmax = fmax
        self.mel_bins = mel_bins
        self.ref = ref
        self.amin = amin
        self.top_db = top_db
        super().__init__(order=order)
        self._spectrogram = torchlibrosa.stft.Spectrogram(
            n_fft=self.window_size,
            win_length=self.window_size,
            hop_length=self.hop_size,
        )
        self._mel = torchlibrosa.stft.LogmelFilterBank(
            sr=self.sample_rate,
            fmin=self.fmin,
            fmax=self.fmax,
            n_mels=self.mel_bins,
            n_fft=self.window_size,
            ref=self.ref,
            amin=self.amin,
            top_db=self.top_db,
        )

    def __call__(self, data: torch.Tensor) -> torch.Tensor:
        return self._mel(self._spectrogram(data)).squeeze(1)


class Resize(AbstractTransform):
    def __init__(
        self,
        height: int = 64,
        width: int = 64,
        interpolation: str = "bilinear",
        antialias: bool = True,
        order: int = -90,
    ) -> None:
        """Resize an image to a specific height and width.
        If "Any" is provided for the height or width, the other dimension will
        be used as the size, creating a square image.

        Args:
            height: The target height. If set to "Any", the width will be used
                as the size. Defaults to 64.
            width: The target width. If set to "Any", the height will be used
                as the size. Defaults to 64.
            interpolation: The interpolation method to use. Defaults to
                'bilinear'.
            antialias: Whether to use antialiasing. Defaults to True.
            order: The order of the transform in the pipeline. Defaults to -90.
        """
        super().__init__(order=order)
        self.width = width
        self.height = height
        self.antialias = antialias
        self.interpolation = interpolation
        if self.width == "Any" and self.height == "Any":
            raise ValueError(
                "At least one of 'height' or 'width' must be specified."
            )
        if self.height == "Any":
            self._size = self.width
        elif self.width == "Any":
            self._size = self.height
        else:
            self._size = (self.height, self.width)
        self._resize = T.Resize(
            size=self._size,
            interpolation=getattr(
                T.InterpolationMode, self.interpolation.upper()
            ),
            antialias=self.antialias,
        )

    def __call__(self, data: torch.Tensor) -> torch.Tensor:
        return self._resize(data)


class SquarePadCrop(AbstractTransform):
    def __init__(self, mode: str = "pad", order: int = -91) -> None:
        """Pad or crop an image to make it square.
        If the image is padded, the padding will be added to the shorter
        sides as black pixels.

        Args:
            mode: The mode to use in ["pad", "crop"]. Defaults to "pad".
            order: The order of the transform in the pipeline. Defaults to -91.

        Raises:
            ValueError: If an invalid mode is provided.
        """
        super().__init__(order=order)
        self.mode = mode
        if self.mode == "pad":
            self._mode_fn = max
        elif self.mode == "crop":
            self._mode_fn = min
        else:
            raise ValueError(f"Invalid mode '{self.mode}'.")

    def __call__(self, data: torch.Tensor) -> torch.Tensor:
        return T.functional.center_crop(data, self._mode_fn(data.shape[-2:]))


class ScaleRange(AbstractTransform):
    def __init__(
        self,
        range: List[float] = [0.0, 1.0],
        order: int = 90,
    ) -> None:
        """Scale a tensor to a specific target range.

        Args:
            range: The range to scale the tensor to. Defaults to [0.0, 1.0].
            order: The order of the transform in the pipeline. Defaults to 90.
        """
        super().__init__(order=order)
        if len(range) != 2:
            raise ValueError(
                f"Expected 'range' to be a list of length 2 but got {range}."
            )
        if range[0] == range[1]:
            raise ValueError(
                "The lower bound of 'range' must be different from the upper "
                f"bound but got {range}."
            )
        self.range = sorted(range)

    def __call__(self, data: torch.Tensor) -> torch.Tensor:
        data_min, data_max = data.min(), data.max()
        if data_min == data_max:
            return torch.full_like(data, data_min)
        data = (data - data_min) / (data_max - data_min)
        return data * (self.range[1] - self.range[0]) + self.range[0]


class Normalize(AbstractTransform):
    def __init__(
        self,
        mean: List[float],
        std: List[float],
        order: int = 95,
    ) -> None:
        """Normalize a tensor with a mean and standard deviation.

        Args:
            mean: The mean to use for normalization.
            std: The standard deviation to use for normalization.
            order: The order of the transform in the pipeline. Defaults to 95.
        """
        super().__init__(order=order)
        self.mean = mean
        self.std = std
        self._normalize = T.Normalize(mean=mean, std=std)

    def __call__(self, data: torch.Tensor) -> torch.Tensor:
        if data.dtype == torch.uint8:
            data = data.float()
        return self._normalize(data)


class FeatureExtractor(AbstractTransform):
    def __init__(
        self,
        fe_type: Optional[str] = None,
        fe_transfer: Optional[str] = None,
        sampling_rate: int = 16000,
        order: int = -80,
    ) -> None:
        """Extract features from an audio signal using a feature extractor
        from the Hugging Face Transformers library.

        Args:
            fe_type: The class of feature extractor to use in ["AST", "Whisper",
                "W2V2", None]. If None, the AutoFeatureExtractor will be used.
                Defaults to None.
            fe_transfer: The name of a pretrained feature extractor to use.
                If None, the feature extractor will be initialized with default
                values. Defaults to None.
            sampling_rate: The sampling rate of the audio signal. Defaults to
                16000.
            order: The order of the transform in the pipeline. Defaults to -80.

        Raises:
            ValueError: If neither 'fe_type' nor 'fe_transfer' is provided.
        """
        super().__init__(order=order)
        if fe_type is None and fe_transfer is None:
            raise ValueError(
                "Either 'fe_type' or 'fe_transfer' must be provided."
            )
        self.fe_type = fe_type
        self.fe_transfer = fe_transfer
        self.sampling_rate = sampling_rate
        fe_class = FE_MAPPINGS[self.fe_type]["fe"]
        padding = FE_MAPPINGS[self.fe_type]["padding"]

        if self.fe_transfer is not None:
            feature_extractor = fe_class.from_pretrained(self.fe_transfer)
        else:
            feature_extractor = fe_class()
            extractor_dict = {
                k: repr(v) for k, v in feature_extractor.__dict__.items()
            }
            warnings.warn(
                f"{fe_class.__name__} "
                "initialized with default values:\n"
                f"{OmegaConf.to_yaml(extractor_dict)}"
            )

        def extract_features(signal: np.ndarray) -> torch.Tensor:
            if len(signal.shape) == 2:
                signal = signal.mean(0)
            extracted = feature_extractor(
                signal,
                sampling_rate=self.sampling_rate,
                padding=padding,
                return_tensors="pt",
            )
            return extracted[list(extracted.keys())[0]][0]

        self._extract_features = extract_features

    def __call__(self, data: torch.Tensor) -> torch.Tensor:
        return self._extract_features(data.numpy())


class Expand(AbstractTransform):
    def __init__(
        self,
        size: int,
        method: str = "pad",
        axis: int = -2,
        order: int = -85,
    ) -> None:
        """Expand a tensor along a specific axis to a specific size, ensuring
        it is at least the specified size.
        If the tensor is smaller than the target size, it will be padded
        with zeros.
        If the tensor is larger than the target size, it will not be cropped.

        Args:
            size: The target size.
            method: The method to use in ["pad", "replicate"]. Defaults to
                "pad".
            axis: The axis along which to expand. Defaults to -2.
            order: The order of the transform in the pipeline. Defaults to -85.
        """
        super().__init__(order=order)
        self.size = size
        self.method = method
        self.axis = axis
        self._expand = AT.Expand(
            size=self.size, method=self.method, axis=self.axis
        )
        self._to_tensor = AnyToTensor()

    def __call__(self, data: torch.Tensor) -> torch.Tensor:
        return self._to_tensor(self._expand(data))


class RandomCrop(AbstractTransform):
    def __init__(
        self,
        size: int,
        method: str = "pad",
        axis: int = -2,
        fix_randomization: bool = False,
        order: int = -90,
    ) -> None:
        """Randomly crop a tensor along a specific axis to a specific size,
        ensuring it is the specified size.

        If the tensor is larger than the target size, it will be randomly
        cropped along the specified axis.

        If the tensor is smaller than the target size, it will be padded.

        Args:
            size: The target size.
            method: Padding method to use if the tensor is smaller than the
                target size in ["pad", "replicate"]. Defaults to "pad".
            axis: The axis along which to crop. Defaults to -2.
            fix_randomization: Whether to fix the randomization. Defaults to
                False.
            order: The order of the transform in the pipeline. Defaults to -90.
        """
        super().__init__(order=order)
        self.size = size
        self.method = method
        self.axis = axis
        self.fix_randomization = fix_randomization
        self._random_crop = AT.RandomCrop(
            size=self.size,
            method=self.method,
            axis=self.axis,
            fix_randomization=self.fix_randomization,
        )
        self._to_tensor = AnyToTensor()

    def __call__(self, data: torch.Tensor) -> torch.Tensor:
        return self._to_tensor(self._random_crop(data))


class StereoToMono(AbstractTransform):
    def __init__(self, order: int = -95) -> None:
        """Convert a stereo audio signal to mono by taking the mean of the
        first dimension.

        Args:
            order: The order of the transform in the pipeline. Defaults to -95.
        """
        super().__init__(order=order)

    def __call__(self, data: torch.Tensor) -> torch.Tensor:
        return data.mean(0, keepdim=True)


class RGBAToRGB(AbstractTransform):
    """Convert an RGBA image to RGB by dropping the alpha channel.

    Args:
        order: The order of the transform in the pipeline. Defaults to -95.
    """

    def __init__(self, order: int = -95) -> None:
        super().__init__(order=order)

    def __call__(self, data: torch.Tensor) -> torch.Tensor:
        return data[:3]


class RGBToGrayscale(AbstractTransform):
    def __init__(self, order: int = 100) -> None:
        """Convert an RGB image to a grayscale image.

        For the conversion to grayscale, the
        luminance is calculated in line with the implementation in
        torchvision.transforms.Grayscale:
        Y = 0.2989 * R + 0.587 * G + 0.114 * B.

        Args:
            order: The order of the transform in the pipeline. Defaults to 100.
        """
        super().__init__(order=order)
        self._to_grayscale = T.Grayscale()

    def __call__(self, image: Union[torch.Tensor, np.ndarray]) -> torch.Tensor:
        image_grs = self._to_grayscale(image)
        return image_grs


class RGBAToGrayscale(AbstractTransform):
    def __init__(self, order: int = -95) -> None:
        """Convert an RGBA image to grayscale by dropping the alpha channel
        and converting to grayscale.

        For the conversion to grayscale, the
        luminance is calculated in line with the implementation in
        torchvision.transforms.Grayscale:
        Y = 0.2989 * R + 0.587 * G + 0.114 * B.


        Args:
            order: The order of the transform in the pipeline. Defaults to -95.
        """
        super().__init__(order=order)
        self._to_grayscale = T.Grayscale()

    def __call__(self, data: torch.Tensor) -> torch.Tensor:
        return self._to_grayscale(data[:3])


class GrayscaleToRGB(AbstractTransform):
    def __init__(self, order: int = 100) -> None:
        """Convert a grayscale image to an RGB image.

        Args:
            order: The order of the transform in the pipeline. Defaults to 100.
        """
        super().__init__(order=order)

    def __call__(self, image: Union[torch.Tensor, np.ndarray]) -> torch.Tensor:
        image = _to_numpy(image)
        image_rgb = AT.Upmix(3, axis=0)(image)
        return _to_tensor(image_rgb)


class Resample(AbstractTransform):
    def __init__(
        self,
        current_sr: int,
        target_sr: int,
        order: int = -95,
    ) -> None:
        """Resample an audio signal to a target sample rate.

        Args:
            current_sr: The current sample rate.
            target_sr: The target sample rate.
            order: The order of the transform in the pipeline. Defaults to -95.
        """
        self.current_sr = current_sr
        self.target_sr = target_sr
        super().__init__(order=order)
        self._resample = TT.Resample(
            orig_freq=self.current_sr, new_freq=self.target_sr
        )

    def __call__(self, data: torch.Tensor) -> torch.Tensor:
        return self._resample(data)


class OpenSMILE(AbstractTransform):
    def __init__(
        self,
        feature_set: str,
        sample_rate: int,
        functionals: bool = False,
        order: int = -80,
    ) -> None:
        """Extract features from an audio signal using openSMILE.

        The openSMILE library must be installed to use this transform.
        To install the required extras, run:
        'pip install autrainer[opensmile]'.

        Args:
            feature_set: The feature set to use.
            sample_rate: The sample rate of the audio signal.
            functionals: Whether to use functionals. Defaults to False.
            order: The order of the transform in the pipeline. Defaults to -80.

        Raises:
            ImportError: If openSMILE is not available.
        """
        if not OPENSMILE_AVAILABLE:
            raise ImportError(
                "openSMILE is not available for feature extraction. Install "
                "the required extras with 'pip install autrainer[opensmile]'."
            )  # pragma: no cover
        super().__init__(order=order)
        self.feature_set = feature_set
        self.sample_rate = sample_rate
        self.functionals = functionals
        if self.functionals:
            self.smile = opensmile.Smile(self.feature_set)
        else:
            self.smile = opensmile.Smile(
                self.feature_set,
                feature_level=opensmile.FeatureLevel.LowLevelDescriptors,
            )

    def __call__(self, data: torch.Tensor) -> torch.Tensor:
        data = self.smile.process_signal(data.numpy(), self.sample_rate)
        return torch.from_numpy(self.smile.to_numpy(data))



import os
from pathlib import Path
from typing import Dict, List, Union
import warnings

from omegaconf import DictConfig

from autrainer.metrics import AbstractMetric

from .abstract_logger import (
    EXPORT_ARTIFACTS,
    AbstractLogger,
    get_params_to_export,
)
from .fallback_logger import FallbackLogger


try:
    with warnings.catch_warnings():
        warnings.filterwarnings("ignore", category=DeprecationWarning)
        import mlflow

    MLFLOW_AVAILABLE = True
except ImportError:  # pragma: no cover
    MLFLOW_AVAILABLE = False  # pragma: no cover


class MLFlowLogger(AbstractLogger):
    def __init__(
        self,
        exp_name: str,
        run_name: str,
        metrics: List[AbstractMetric],
        tracking_metric: AbstractMetric,
        artifacts: List[Union[str, Dict[str, str]]] = EXPORT_ARTIFACTS,
        output_dir: str = "mlruns",
    ) -> None:
        super().__init__(
            exp_name, run_name, metrics, tracking_metric, artifacts
        )
        output_dir = Path(output_dir)
        if not output_dir.is_absolute():
            output_dir = output_dir.absolute()
        if not any(
            str(output_dir).startswith(prefix)
            for prefix in ["file://", "http://", "https://"]
        ):
            output_dir = output_dir.as_uri()
        self.output_dir = output_dir

    def setup(self) -> None:
        mlflow.set_tracking_uri(self.output_dir)
        self.exp_id = self._get_or_create_experiment()
        mlflow.set_experiment(experiment_id=self.exp_id)
        self.run = self._get_or_create_run()

    def _get_or_create_experiment(self) -> str:
        experiment = mlflow.get_experiment_by_name(self.exp_name)
        if experiment:
            return experiment.experiment_id
        return mlflow.create_experiment(name=self.exp_name)

    def _get_or_create_run(self) -> "mlflow.ActiveRun":
        self._delete_run_if_exists(self.run_name)
        run = mlflow.start_run(run_name=self.run_name)
        return run

    def _delete_run_if_exists(self, run_name: str) -> None:
        client = mlflow.MlflowClient()
        runs = mlflow.search_runs(
            experiment_ids=[self.exp_id],
            filter_string=f"tags.mlflow.runName='{run_name}'",
        )
        if runs.shape[0] > 0:
            run_id = runs.iloc[0]["run_id"]
            client.delete_run(run_id)

    def log_params(self, params: Union[dict, DictConfig]) -> None:
        params = get_params_to_export(params)
        mlflow.log_params(params)

    def log_metrics(
        self,
        metrics: Dict[str, Union[int, float]],
        iteration=None,
    ) -> None:
        mlflow.log_metrics(metrics, step=iteration)

    def log_timers(self, timers: Dict[str, float]) -> None:
        mlflow.log_params(timers)

    def log_artifact(self, filename: str, path: str = "") -> None:
        mlflow.log_artifact(os.path.join(path, filename))

    def end_run(self) -> None:
        mlflow.end_run()


MLFlowLogger = (
    MLFlowLogger
    if MLFLOW_AVAILABLE
    else lambda *args, **kwargs: FallbackLogger(
        "MLFlowLogger", "mlflow"
    )  # pragma: no cover
)



from typing import Dict, List, Optional, Tuple, Union

from omegaconf import DictConfig, OmegaConf

import autrainer

from .abstract_transform import AbstractTransform
from .smart_compose import SmartCompose
from .specific_transforms import GrayscaleToRGB, RGBToGrayscale


class TransformManager:
    def __init__(
        self,
        model_transform: Union[DictConfig, Dict],
        dataset_transform: Union[DictConfig, Dict],
        train_augmentation: Optional[SmartCompose] = None,
        dev_augmentation: Optional[SmartCompose] = None,
        test_augmentation: Optional[SmartCompose] = None,
    ) -> None:
        """Transform manager for composing transforms for the train, dev, and
        test datasets.
        Automatically handles the creation of all transformation pipelines
        including incorporating optional augmentations.

        Args:
            model_transform: The model transform configuration.
            dataset_transform: The dataset transform configuration.
            train_augmentation: The train augmentation pipeline.
                Defaults to None.
            dev_augmentation: The dev augmentation pipeline.
                Defaults to None.
            test_augmentation: The test augmentation pipeline.
                Defaults to None.
        """
        if isinstance(model_transform, DictConfig):
            model_transform = OmegaConf.to_container(model_transform)
        if isinstance(dataset_transform, DictConfig):
            dataset_transform = OmegaConf.to_container(dataset_transform)
        self.model_transform = model_transform
        self.dataset_transform = dataset_transform
        self.train_augmentation = train_augmentation
        self.dev_augmentation = dev_augmentation
        self.test_augmentation = test_augmentation

    def get_transforms(
        self,
    ) -> Tuple[SmartCompose, SmartCompose, SmartCompose]:
        """Get the composed transform pipelines for the train, dev, and test
        datasets.

        Returns:
            The composed transform pipelines for the train, dev, and test
                datasets.
        """
        base = self._specific("base") + self._match_model_dataset()

        train = base + self._specific("train") + self.train_augmentation
        dev = base + self._specific("dev") + self.dev_augmentation
        test = base + self._specific("test") + self.test_augmentation
        return train, dev, test

    def _specific(self, subset: str) -> SmartCompose:
        """Get and build the specific transforms for the model and dataset
        specified in the transform configurations.

        Args:
            subset: The transform subset to get. Must be one of "train", "dev",
                "test", or "base".

        Raises:
            ValueError: If the subset is not one of "train", "dev", "test",
                or "base".

        Returns:
            The composed transform pipeline.
        """
        if subset not in ["train", "dev", "test", "base"]:
            raise ValueError(
                "Transform subset must be one of 'train', 'dev', 'test', "
                f"or 'base', got {subset}."
            )
        model_transforms = self.model_transform.get(subset, [])
        dataset_transforms = self.dataset_transform.get(subset, [])
        if not model_transforms and not dataset_transforms:
            return SmartCompose([])
        model_transforms, dataset_transforms = self._filter_transforms(
            model_transforms, dataset_transforms
        )
        combined_transforms = self._combine_transforms(
            model_transforms,
            dataset_transforms,
        )
        transforms = [
            autrainer.instantiate_shorthand(t, instance_of=AbstractTransform)
            for t in combined_transforms
        ]
        return SmartCompose(transforms)

    def _filter_transforms(
        self,
        model_transforms: List[Union[str, Dict]],
        dataset_transforms: List[Union[str, Dict]],
    ) -> Tuple[List[Union[str, Dict]], List[Union[str, Dict]]]:
        """Remove None values from the model and dataset transforms.
        None values are used to indicate that the transform was removed
        from the config by either the model or dataset transform.
        See `Automatic Transforms` for more details.

        Args:
            model_transforms: List of model transforms.
            dataset_transforms: List of dataset transforms.

        Returns:
            The filtered model and dataset transforms.
        """

        def _filter(l1, l2):
            for transform in l1.copy():
                if isinstance(transform, str):
                    continue
                key = next(iter(transform.keys()))
                if transform[key] is None and len(transform.keys()) == 1:
                    l1.remove(transform)
                    d_idx = self._find_matching_transform(l2, key)
                    if d_idx is not None:
                        l2.pop(d_idx)
            return l1, l2

        model_transforms, dataset_transforms = _filter(
            model_transforms, dataset_transforms
        )
        dataset_transforms, model_transforms = _filter(
            dataset_transforms, model_transforms
        )
        return model_transforms, dataset_transforms

    def _combine_transforms(
        self,
        model_transforms: List[Union[str, Dict]],
        dataset_transforms: List[Union[str, Dict]],
    ) -> List[Union[str, Dict]]:
        """Combine model and dataset transforms into a single list.
        Model transforms outweigh dataset transforms, so if a transform
        is present in both lists, the model transform will be used.

        Args:
            model_transforms: List of model transforms.
            dataset_transforms: List of dataset transforms.

        Returns:
            The combined list of transforms.
        """
        combined = model_transforms
        for transform in dataset_transforms:
            if isinstance(transform, str):
                if transform not in combined:
                    combined.append(transform)
            else:
                key = next(iter(transform.keys()))
                if self._find_matching_transform(combined, key) is None:
                    combined.append(transform)
        return combined

    @staticmethod
    def _find_matching_transform(
        l: List[Union[str, Dict]],
        key: str,
    ) -> Optional[int]:
        """Find the index of a transform in a list of transforms.
        If the transform is a dictionary, the key is used to find the
        transform. If the transform is a string, the string itself is
        used to find the transform.

        Args:
            l: List of transforms.
            key: The key to search for.

        Returns:
            The index of the transform in the list. If the transform is not
                found, None is returned.
        """
        x = (
            idx
            for idx, t in enumerate(l)
            if (isinstance(t, dict) and key in t.keys()) or t == key
        )
        return next(x, None)

    def _match_model_dataset(self) -> SmartCompose:
        """Match the model input type with the dataset type by adding
        the necessary transform.

        Raises:
            ValueError: If the model and dataset types do not match.

        Returns:
            The composed transform to match the model and dataset types.
        """
        m = self.model_transform["type"]
        d = self.dataset_transform["type"]
        match = []
        if m == "image" and d == "grayscale":
            match = [GrayscaleToRGB()]
        elif m == "grayscale" and d == "image":
            match = [RGBToGrayscale()]
        elif m != d:
            raise ValueError(
                f"Model input type '{m}' does not match dataset type '{d}'"
            )
        return SmartCompose(match)



from dataclasses import dataclass
import os

from .abstract_script import AbstractScript
from .command_line_error import CommandLineError


@dataclass
class AbstractPostprocessArgs:
    results_dir: str
    experiment_id: str


class AbstractPostprocessScript(AbstractScript):
    def add_arguments(self) -> None:
        self.parser.add_argument(
            "results_dir",
            type=str,
            help="Path to grid search results directory.",
        )
        self.parser.add_argument(
            "experiment_id",
            type=str,
            help="ID of experiment to postprocess.",
        )

    def main(self, args: AbstractPostprocessArgs) -> None:
        self._assert_results_dir_exists(args)
        self._assert_experiment_id_exists(args)

    def _assert_results_dir_exists(
        self, args: AbstractPostprocessArgs
    ) -> None:
        if os.path.exists(args.results_dir) and os.path.isdir(
            args.results_dir
        ):
            return
        raise CommandLineError(
            self.parser,
            f"Results directory '{args.results_dir}' does not exist.",
            code=1,
        )

    def _assert_experiment_id_exists(
        self, args: AbstractPostprocessArgs
    ) -> None:
        path = os.path.join(args.results_dir, args.experiment_id)
        if os.path.exists(path) and os.path.isdir(path):
            return
        raise CommandLineError(
            self.parser,
            (
                f"Experiment ID '{args.experiment_id}' does not exist "
                f"in the results directory '{args.results_dir}'."
            ),
            code=1,
        )



import os
from pathlib import Path
import shutil
from typing import Dict, List, Optional, Union

from omegaconf import DictConfig
import pandas as pd
from torchaudio import datasets

from autrainer.transforms import SmartCompose

from .abstract_dataset import BaseClassificationDataset


class SpeechCommands(BaseClassificationDataset):
    def __init__(
        self,
        path: str,
        features_subdir: str,
        seed: int,
        metrics: List[Union[str, DictConfig, Dict]],
        tracking_metric: Union[str, DictConfig, Dict],
        index_column: str,
        target_column: Union[str, List[str]],
        file_type: str,
        file_handler: Union[str, DictConfig, Dict],
        batch_size: int,
        inference_batch_size: Optional[int] = None,
        train_transform: Optional[SmartCompose] = None,
        dev_transform: Optional[SmartCompose] = None,
        test_transform: Optional[SmartCompose] = None,
        stratify: Optional[List[str]] = None,
    ) -> None:
        """Speech Commands (v0.02) dataset.

        Args:
            path: Root path to the dataset.
            features_subdir: Subdirectory containing the features.
            seed: Seed for reproducibility.
            metrics: List of metrics to calculate.
            tracking_metric: Metric to track.
            index_column: Index column of the dataframe.
            target_column: Target column of the dataframe.
            file_type: File type of the features.
            file_handler: File handler to load the data.
            batch_size: Batch size.
            inference_batch_size: Inference batch size. If None, defaults to
                batch_size. Defaults to None.
            train_transform: Transform to apply to the training set.
                Defaults to None.
            dev_transform: Transform to apply to the development set.
                Defaults to None.
            test_transform: Transform to apply to the test set.
                Defaults to None.
            stratify: Columns to stratify the dataset on. Defaults to None.
        """
        super().__init__(
            path=path,
            features_subdir=features_subdir,
            seed=seed,
            metrics=metrics,
            tracking_metric=tracking_metric,
            index_column=index_column,
            target_column=target_column,
            file_type=file_type,
            file_handler=file_handler,
            batch_size=batch_size,
            inference_batch_size=inference_batch_size,
            train_transform=train_transform,
            dev_transform=dev_transform,
            test_transform=test_transform,
            stratify=stratify,
        )

    @staticmethod
    def download(path: str) -> None:  # pragma: no cover
        """Download the Speech Commands (v0.02) dataset from torchaudio.

        For more information on the dataset, see:
        https://doi.org/10.48550/arXiv.1804.03209

        Args:
            path: Path to the directory to download the dataset to.
        """
        if os.path.isdir(os.path.join(path, "default")):
            return

        _version = "speech_commands_v0.02"
        subsets = ["training", "validation", "testing"]
        df_names = ["train", "dev", "test"]
        os.makedirs(path, exist_ok=True)
        datasets.SPEECHCOMMANDS(
            root=path,
            download=True,
            folder_in_archive=_version,
        )

        for subset, df_name in zip(subsets, df_names):
            dataset = datasets.SPEECHCOMMANDS(
                root=path,
                subset=subset,
                download=False,
                folder_in_archive=_version,
            )
            metadata = []
            for n in range(len(dataset)):
                meta = dataset.get_metadata(n)
                metadata.append(
                    {
                        "path": Path(meta[0]).relative_to(_version),
                        "sample_rate": meta[1],
                        "label": meta[2],
                        "speaker_id": meta[3],
                        "utterance_number": meta[4],
                    }
                )
            pd.DataFrame(metadata).to_csv(
                os.path.join(path, f"{df_name}.csv"),
                index=False,
            )

        shutil.move(
            os.path.join(path, _version, _version),
            os.path.join(path, "default"),
        )
        shutil.rmtree(os.path.join(path, _version))



import copy
from typing import List

import numpy as np
from omegaconf import DictConfig
import pandas as pd
import torch

from autrainer.core.utils import Bookkeeping
from autrainer.datasets import AbstractDataset


class OutputsTracker:
    def __init__(
        self,
        export: bool,
        prefix: str,
        data: AbstractDataset,
        criterion: DictConfig,
        bookkeeping: Bookkeeping = None,
    ) -> None:
        self._export = export
        self._prefix = prefix
        self._data = data
        self._bookkeeping = bookkeeping
        self._tracker_criterion = copy.deepcopy(criterion).cpu()
        self._tracker_criterion.reduction = "none"
        self.reset()

    def reset(self) -> None:
        self._outputs = torch.zeros(
            (0, self._data.output_dim), dtype=torch.float32
        )
        self._targets = torch.zeros(0, dtype=torch.long)
        self._indices = torch.zeros(0, dtype=torch.long)
        self._losses = torch.zeros(0, dtype=torch.float32)
        self._predictions = None
        self._results_df = None

    def update(
        self,
        output: torch.Tensor,
        target: torch.Tensor,
        sample_idx: torch.Tensor,
    ) -> None:
        output = output.cpu()
        target = target.cpu()
        self._outputs = torch.cat([self._outputs, output], dim=0)
        self._targets = torch.cat([self._targets, target], dim=0)
        self._indices = torch.cat([self._indices, sample_idx], dim=0)

        with torch.no_grad():
            loss = self._tracker_criterion(output, target)
            self._losses = torch.cat([self._losses, loss], dim=0)

    def save(self, iteration_folder: str, reset=True) -> None:
        results = {
            "outputs": self._outputs.numpy(),
            "targets": self._targets.numpy(),
            "indices": self._indices.numpy(),
            "losses": self._losses.numpy(),
        }

        self._predictions = self._data.target_transform.predict_batch(
            self._outputs
        )
        self._results_df = pd.DataFrame(index=results["indices"])
        self._results_df["predictions"] = self._predictions
        self._results_df["predictions"] = self._results_df[
            "predictions"
        ].apply(self._data.target_transform.decode)

        if self._export:
            for key, value in results.items():
                self._bookkeeping.save_results_np(
                    value, f"{self._prefix}_{key}.npy", iteration_folder
                )
            self._bookkeeping.save_results_df(
                self._results_df.reset_index(),
                f"{self._prefix}_results.csv",
                iteration_folder,
            )

        if reset:
            self.reset()

    def check_saved(func):
        def wrapper(self, *args, **kwargs):
            if self._results_df is None:
                raise ValueError("Results not saved yet.")
            return func(self, *args, **kwargs)

        return wrapper

    @property
    @check_saved
    def outputs(self) -> np.ndarray:
        return self._outputs.numpy()

    @property
    @check_saved
    def targets(self) -> np.ndarray:
        return self._targets.numpy()

    @property
    @check_saved
    def indices(self) -> np.ndarray:
        return self._indices.numpy()

    @property
    @check_saved
    def losses(self) -> np.ndarray:
        return self._losses.numpy()

    @property
    @check_saved
    def predictions(self) -> np.ndarray:
        return self._predictions

    @property
    @check_saved
    def results_df(self) -> pd.DataFrame:
        return self._results_df


def init_trackers(
    exports: List[bool],
    prefixes: List[str],
    data: AbstractDataset,
    criterion: DictConfig,
    bookkeeping: Bookkeeping = None,
) -> OutputsTracker:
    trackers = []
    for export, prefix in zip(exports, prefixes):
        trackers.append(
            OutputsTracker(
                export=export,
                prefix=prefix,
                data=data,
                criterion=criterion,
                bookkeeping=bookkeeping,
            )
        )

    return trackers



import os
import shutil
from typing import Dict, List, Optional, Union

from omegaconf import DictConfig, OmegaConf
import pandas as pd

import autrainer
from autrainer.core.constants import NAMING_CONVENTION
from autrainer.core.plotting import PlotMetrics
from autrainer.loggers import get_params_to_export
from autrainer.metrics import AbstractMetric

from .postprocessing_utils import (
    get_plotting_params,
    get_run_names,
    get_training_type,
    load_yaml,
    save_yaml,
)


class SummarizeGrid:
    def __init__(
        self,
        results_dir: str,
        experiment_id: str,
        summary_dir: str = "summary",
        training_dir: str = "training",
        clear_old_outputs: bool = True,
        training_type: Optional[str] = None,
        max_runs_plot: Optional[int] = None,
        plot_params: Optional[Union[DictConfig, Dict]] = None,
    ) -> None:
        """Summarize the results of a grid search.

        Args:
            results_dir: The directory where the results are stored.
            experiment_id: The ID of the grid search experiment.
            summary_dir: The directory where the the grid search summary will
                be stored. Defaults to "summary".
            training_dir: The directory of the training results of the
                    experiment. Defaults to "training".
            clear_old_outputs: Whether to clear existing summary outputs.
                Defaults to True.
            training_type: The type of training in ["Epoch", "Step"]. If None,
                it will be inferred from the training results. Defaults to None.
            max_runs_plot: The maximum number of best runs to plot. If None,
                all runs will be plotted. Defaults to None.
            plot_params: Additional parameters for plotting. Defaults to None.
        """
        self.results_dir = results_dir
        self.experiment_id = experiment_id
        self.output_directory = os.path.join(
            self.results_dir, self.experiment_id, summary_dir
        )
        self.training_directory = os.path.join(
            self.results_dir, self.experiment_id, training_dir
        )
        if os.path.exists(self.output_directory) and clear_old_outputs:
            shutil.rmtree(self.output_directory)
        os.makedirs(self.output_directory, exist_ok=True)

        self.run_names = get_run_names(self.training_directory)
        self.training_type = training_type
        if self.training_type is None:
            self.training_type = get_training_type(
                self.training_directory, self.run_names
            )
        self.max_runs_plot = max_runs_plot
        self.plot_params = plot_params
        self.run_names.sort()
        if self.plot_params is None:
            self.plot_params = get_plotting_params(
                self.training_directory, self.run_names[0]
            )
        self.metric_fns: List[AbstractMetric] = []

    def summarize(self) -> None:
        """Summarize the results of the grid search."""
        self._summarize_metrics()
        self._summarize_config()
        self._summarize_best_runs()

    def _summarize_metrics(self) -> None:
        records = []
        for n in self.run_names:
            metrics = self._read_metrics(n)
            test = self._read_test_metrics(n)
            records.append({"run_name": n, **metrics, **test})
        df = pd.DataFrame.from_records(records)
        path = os.path.join(self.output_directory, "metrics.csv")
        df.to_csv(path, index=False)

    def _summarize_config(self) -> None:
        records = []
        for n in self.run_names:
            run_dir = os.path.join(self.training_directory, n)
            config_path = os.path.join(run_dir, ".hydra", "config.yaml")
            config = load_yaml(config_path)
            config = dict(sorted(get_params_to_export(config).items()))
            time = self._read_times(n)
            records.append({"run_name": n, **config, **time})
        df = pd.DataFrame.from_records(records)
        path = os.path.join(self.output_directory, "config.csv")
        df.to_csv(path, index=False)

    def _summarize_best_runs(self) -> None:
        summary_path = os.path.join(self.output_directory, "metrics.csv")
        df = pd.read_csv(summary_path)
        metric_cols = [m for m in df.columns if not m.endswith(".std")]
        metric_cols.remove("run_name")
        metric_cols.remove("best_iteration")
        best_runs = {}
        for metric in metric_cols:
            if "loss" in metric:
                idx = df[metric].idxmin()
            else:
                base_metric = metric.replace("test_", "").replace("train_", "")
                m = next(
                    (fn for fn in self.metric_fns if fn.name == base_metric),
                    None,
                )
                if m is None:
                    raise ValueError(f"No metric found for '{metric}'.")
                idx = m.get_best_pos(df[metric])
            best_runs[metric] = df.loc[idx, "run_name"]
        path = os.path.join(self.output_directory, "best_runs.yaml")
        save_yaml(path, best_runs)

    def _find_metrics_to_plot(self) -> List[str]:
        summary_path = os.path.join(self.output_directory, "metrics.csv")
        df = pd.read_csv(summary_path)
        plot_metrics = [
            m
            for m in list(df.columns)
            if m in [fn.name for fn in self.metric_fns]
            or m in ["train_loss", "dev_loss"]
        ]
        return plot_metrics

    def plot_metrics(self) -> None:
        """Plot the metrics of the grid search."""
        plotter = PlotMetrics(
            output_directory=self.output_directory,
            training_type=self.training_type,
            metric_fns=self.metric_fns,
            **self.plot_params,
        )
        metrics_to_plot = self._find_metrics_to_plot()
        for metric in metrics_to_plot:
            metrics_list = []
            metrics_std_list = []
            for n in self.run_names:
                metrics_path = os.path.join(
                    self.training_directory, n, "metrics.csv"
                )
                df = pd.read_csv(metrics_path, index_col="iteration")
                if metric not in df.columns:
                    continue
                metrics_list.append(df[[metric]].rename(columns={metric: n}))

                std_metric = f"{metric}.std"
                if std_metric in df.columns:
                    metrics_std_list.append(
                        df[[std_metric]].rename(columns={std_metric: n})
                    )

            metrics_df = pd.concat(metrics_list, axis=1)
            metrics_std_df = (
                pd.concat(metrics_std_list, axis=1)
                if metrics_std_list
                else None
            )
            plotter.plot_metric(
                metrics_df, metric, metrics_std_df, max_runs=self.max_runs_plot
            )

    def plot_aggregated_bars(self) -> None:
        """Plot additional aggregated bars for the grid search."""
        metrics_to_plot = self._find_metrics_to_plot()
        plotter = PlotMetrics(
            output_directory=self.output_directory,
            training_type=self.training_type,
            metric_fns=self.metric_fns,
            **self.plot_params,
        )
        metrics_df = pd.DataFrame()

        # Iterate over each CSV file and build up metrics dataframe
        for n in self.run_names:
            metrics_path = os.path.join(
                self.training_directory, n, "metrics.csv"
            )
            df = pd.read_csv(metrics_path)
            last_row = df.iloc[-1:]
            run_details = n.split("_")
            for i, param in enumerate(NAMING_CONVENTION):
                last_row.insert(i, param, run_details[i])
            metrics_df = pd.concat((metrics_df, last_row), ignore_index=True)
        for metric in metrics_to_plot:
            plotter.plot_aggregated_bars(metrics_df, metric)

    def _read_metrics(self, run_name: str) -> dict:
        metrics_path = os.path.join(
            self.training_directory, run_name, "metrics.csv"
        )
        config_path = os.path.join(
            self.training_directory, run_name, ".hydra", "config.yaml"
        )
        df = pd.read_csv(metrics_path, index_col="iteration")
        cfg = OmegaConf.load(config_path)
        tracking_metric = autrainer.instantiate_shorthand(
            config=cfg.dataset.tracking_metric,
            instance_of=AbstractMetric,
        )
        self._update_metric_fns(cfg.dataset.metrics)
        best_iteration = tracking_metric.get_best_pos(df[tracking_metric.name])
        metrics_dict = df.loc[best_iteration].to_dict()
        metrics_dict["best_iteration"] = best_iteration
        return metrics_dict

    def _read_test_metrics(self, run_name: str, prefix: str = "test_") -> dict:
        path = os.path.join(
            self.training_directory, run_name, "_test", "test_holistic.yaml"
        )
        test_dict = load_yaml(path)
        test_dict = {prefix + k: v["all"] for k, v in test_dict.items()}
        return test_dict

    def _read_times(self, run_name: str) -> dict:
        path = os.path.join(self.training_directory, run_name, "timer.yaml")
        time_dict = load_yaml(path)
        return {
            "time_train_mean": time_dict["train"]["mean"],
            "time_dev_mean": time_dict["dev"]["mean"],
            "time_test_mean": time_dict["test"]["mean"],
        }

    def _update_metric_fns(self, metric_cfg: Union[dict, DictConfig]) -> None:
        metrics = [
            autrainer.instantiate_shorthand(m, instance_of=AbstractMetric)
            for m in metric_cfg
        ]
        for m in metrics:
            if m.name in [m.name for m in self.metric_fns]:
                continue
            self.metric_fns.append(m)



from typing import Optional

import torch

import autrainer

from .abstract_augmentation import AbstractAugmentation


try:
    import albumentations  # noqa: F401

    ALBUMENTATIONS_AVAILABLE = True
except ImportError:  # pragma: no cover
    ALBUMENTATIONS_AVAILABLE = False  # pragma: no cover

try:
    import audiomentations  # noqa: F401

    AUDIOMENTATIONS_AVAILABLE = True
except ImportError:  # pragma: no cover
    AUDIOMENTATIONS_AVAILABLE = False  # pragma: no cover

try:
    import torch_audiomentations  # noqa: F401

    TORCHAUDIOMENTATIONS_AVAILABLE = True
except ImportError:  # pragma: no cover
    TORCHAUDIOMENTATIONS_AVAILABLE = False  # pragma: no cover


class AugmentationWrapper(AbstractAugmentation):
    def __init__(
        self,
        augmentation_import_path: AbstractAugmentation,
        order: int = 0,
        p: float = 1.0,
        generator_seed: Optional[int] = None,
        pass_index: bool = False,
        probability_attr: Optional[str] = None,
        **kwargs,
    ) -> None:
        """Wrapper around an AbstractAugmentation.

        Important: While the probability of applying the augmentation is
        deterministic if the generator_seed is set, the actual augmentation
        applied is not deterministic. This is because the internal random
        number generator of the augmentation is not seeded.

        Args:
            augmentation_import_path: The augmentation import path for the
                underlying augmentation to wrap.
            order: The order of the augmentation in the transformation pipeline.
                Defaults to 0.
            p: The probability of applying the augmentation. Defaults to 1.0.
            generator_seed: The initial seed for the internal random number
                generator drawing the probability. If None, the generator is
                not seeded. Defaults to None.
            pass_index: Whether to pass the index to the underlying
                augmentation. Defaults to False.
            probability_attr: Name of the underlying augmentation's
                probability attribute to override with 1.0. If None, the
                probability attribute is not overridden. Defaults to None.
            kwargs: Additional keyword arguments to store in the object.
        """
        super().__init__(order, p, generator_seed, **kwargs)
        self.augmentation_import_path = augmentation_import_path
        self._deterministic = False

        # ? audobject passes _object_root_ to the object, which is not a valid
        # ? argument for the base augmentation class.
        kwargs.pop("_object_root_", None)
        self.augmentation = autrainer.instantiate_shorthand(
            self.augmentation_import_path,
            **kwargs,
        )

        self.pass_index = pass_index
        self.probability_attr = probability_attr
        if self.probability_attr and hasattr(
            self.augmentation, self.probability_attr
        ):
            setattr(self.augmentation, self.probability_attr, 1.0)

    def apply(self, x: torch.Tensor, index: int = None) -> torch.Tensor:
        if self.pass_index:
            return self.augmentation(x, index)  # pragma: no cover
        return self.augmentation(x)


class TorchvisionAugmentation(AugmentationWrapper):
    def __init__(
        self,
        name: str,
        order: int = 0,
        p: float = 1.0,
        generator_seed: Optional[int] = None,
        **kwargs,
    ) -> None:
        """Wrapper around torchvision.transforms.v2 transforms, which are
        specified by their class name and keyword arguments.

        Functionals are currently not supported.

        Important: While the probability of applying the augmentation is
        deterministic if the generator_seed is set, the actual augmentation
        applied is not deterministic. This is because the internal random
        number generator of the augmentation is not seeded.

        Args:
            name: Name of the torchvision augmentation. Must be a valid
                torchvision.transforms.v2 transform class name.
            order: The order of the augmentation in the transformation pipeline.
                Defaults to 0.
            p: The probability of applying the augmentation. Defaults to 1.0.
            generator_seed: The initial seed for the internal random number
                generator drawing the probability. If None, the generator is
                not seeded. Defaults to None.
            kwargs: Keyword arguments passed to the torchvision augmentation.
        """
        self.name = name

        super().__init__(
            augmentation_import_path=f"torchvision.transforms.v2.{self.name}",
            order=order,
            p=p,
            generator_seed=generator_seed,
            pass_index=False,
            probability_attr="p",
            **kwargs,
        )


class TorchaudioAugmentation(AugmentationWrapper):
    def __init__(
        self,
        name: str,
        order: int = 0,
        p: float = 1.0,
        generator_seed: Optional[int] = None,
        **kwargs,
    ) -> None:
        """Wrapper around torchaudio.transforms transforms, which are specified
        by their class name and keyword arguments.

        Important: While the probability of applying the augmentation is
        deterministic if the generator_seed is set, the actual augmentation
        applied is not deterministic. This is because the internal random
        number generator of the augmentation is not seeded.

        Args:
            name: Name of the torchaudio augmentation. Must be a valid
                torchaudio.transforms transform class name.
            order: The order of the augmentation in the transformation pipeline.
                Defaults to 0.
            p: The probability of applying the augmentation. Defaults to 1.0.
            generator_seed: The initial seed for the internal random number
                generator drawing the probability. If None, the generator is
                not seeded. Defaults to None.
            kwargs: Keyword arguments passed to the torchaudio augmentation.
        """
        self.name = name
        super().__init__(
            augmentation_import_path=f"torchaudio.transforms.{self.name}",
            order=order,
            p=p,
            generator_seed=generator_seed,
            pass_index=False,
            probability_attr="p",
            **kwargs,
        )

    def apply(self, x: torch.Tensor, index: int = None) -> torch.Tensor:
        x = super().apply(x, index)
        if isinstance(x, tuple):
            return x[0]
        return x


class AlbumentationsAugmentation(AugmentationWrapper):
    def __init__(
        self,
        name: str,
        order: int = 0,
        p: float = 1.0,
        generator_seed: Optional[int] = None,
        **kwargs,
    ) -> None:
        """Wrapper around albumentations transforms, which are specified by
        their class name and keyword arguments.

        Albumentations operates on numpy arrays, so the input tensor is
        converted to a numpy array before applying the augmentation, and the
        output numpy array is converted back to a tensor.

        Important: While the probability of applying the augmentation is
        deterministic if the generator_seed is set, the actual augmentation
        applied is not deterministic. This is because the internal random
        number generator of the augmentation is not seeded.

        Args:
            name: Name of the albumentations augmentation. Must be a valid
                albumentations transform class name.
            order: The order of the augmentation in the transformation pipeline.
                Defaults to 0.
            p: The probability of applying the augmentation. Defaults to 1.0.
            generator_seed: The initial seed for the internal random number
                generator drawing the probability. If None, the generator is
                not seeded. Defaults to None.
            kwargs: Keyword arguments passed to the albumentations
                augmentation.
        """
        if not ALBUMENTATIONS_AVAILABLE:
            raise ImportError(
                "Albumentations is not installed. Install the required extras "
                "with 'pip install autrainer[albumentations]'."
            )  # pragma: no cover

        self.name = name
        super().__init__(
            augmentation_import_path=f"albumentations.{self.name}",
            order=order,
            p=p,
            generator_seed=generator_seed,
            pass_index=False,
            probability_attr="p",
            **kwargs,
        )

    def apply(self, x: torch.Tensor, index: int = None) -> torch.Tensor:
        device = x.device
        x = x.cpu().permute(1, 2, 0).numpy()
        x = self.augmentation(image=x)["image"]
        x = torch.from_numpy(x).permute(2, 0, 1).to(device)
        return x


class AudiomentationsAugmentation(AugmentationWrapper):
    def __init__(
        self,
        name: str,
        sample_rate: Optional[int] = None,
        order: int = 0,
        p: float = 1.0,
        generator_seed: Optional[int] = None,
        **kwargs,
    ) -> None:
        """Wrapper around audiomentations transforms, which are specified
        by their class name and keyword arguments.

        Audiomentations operates on numpy arrays, so the input tensor is
        converted to a numpy array before applying the augmentation, and the
        output numpy array is converted back to a tensor.

        Important: While the probability of applying the augmentation is
        deterministic if the generator_seed is set, the actual augmentation
        applied is not deterministic. This is because the internal random
        number generator of the augmentation is not seeded.

        Args:
            name: Name of the torchaudio augmentation. Must be a valid
                audiomentations transform class name.
            sample_rate: The sample rate of the audio data. Should be specified
                for most audio augmentations. If None, the sample rate is not
                passed to the augmentation. Defaults to None.
            order: The order of the augmentation in the transformation pipeline.
                Defaults to 0.
            p: The probability of applying the augmentation. Defaults to 1.0.
            generator_seed: The initial seed for the internal random number
                generator drawing the probability. If None, the generator is
                not seeded. Defaults to None.
            kwargs: Keyword arguments passed to the audiomentations
                augmentation.
        """
        if not AUDIOMENTATIONS_AVAILABLE:
            raise ImportError(
                "Audiomentations is not installed. Install the required "
                "extras with 'pip install autrainer[audiomentations]'."
            )  # pragma: no cover

        self.name = name
        self.sample_rate = sample_rate
        self._aug_kwargs = {}
        if self.sample_rate is not None:
            self._aug_kwargs["sample_rate"] = self.sample_rate
        super().__init__(
            augmentation_import_path=f"audiomentations.{self.name}",
            order=order,
            p=p,
            generator_seed=generator_seed,
            pass_index=False,
            probability_attr="p",
            **kwargs,
        )

    def apply(self, x: torch.Tensor, index: int = None) -> torch.Tensor:
        device = x.device
        x = self.augmentation(x.cpu().numpy(), **self._aug_kwargs)
        return torch.from_numpy(x).to(device)


class TorchAudiomentationsAugmentation(AugmentationWrapper):
    def __init__(
        self,
        name: str,
        sample_rate: int,
        order: int = 0,
        p: float = 1.0,
        generator_seed: Optional[int] = None,
        **kwargs,
    ) -> None:
        """Wrapper around torch_audiomentations transforms, which are specified
        by their class name and keyword arguments.

        Important: While the probability of applying the augmentation is
        deterministic if the generator_seed is set, the actual augmentation
        applied is not deterministic. This is because the internal random
        number generator of the augmentation is not seeded.

        Args:
            name: Name of the torchaudio augmentation. Must be a valid
                torch_audiomentations transform class name.
            sample_rate: The sample rate of the audio data.
            order: The order of the augmentation in the transformation pipeline.
                Defaults to 0.
            p: The probability of applying the augmentation. Defaults to 1.0.
            generator_seed: The initial seed for the internal random number
                generator drawing the probability. If None, the generator is
                not seeded. Defaults to None.
            kwargs: Keyword arguments passed to the torch_audiomentations
                augmentation.
        """
        if not TORCHAUDIOMENTATIONS_AVAILABLE:
            raise ImportError(
                "torch-audiomentations is not installed. Install the required "
                "extras with 'pip install autrainer[torch-audiomentations]'."
            )  # pragma: no cover

        self.name = name
        self.sample_rate = sample_rate
        kwargs["output_type"] = "dict"
        super().__init__(
            augmentation_import_path=f"torch_audiomentations.{self.name}",
            order=order,
            p=p,
            generator_seed=generator_seed,
            pass_index=False,
            probability_attr="p",
            **kwargs,
        )

    def apply(self, x: torch.Tensor, index: int = None) -> torch.Tensor:
        x = x.unsqueeze(0)
        x = self.augmentation(x, sample_rate=self.sample_rate).samples
        return x.squeeze(0)



from autrainer.core.scripts import (
    create,
    fetch,
    group,
    inference,
    postprocess,
    preprocess,
    rm_failed,
    rm_states,
    show,
    train,
)
from autrainer.core.scripts import list_configs as list


__all__ = [
    "create",
    "fetch",
    "group",
    "inference",
    "list",
    "postprocess",
    "preprocess",
    "rm_failed",
    "rm_states",
    "show",
    "train",
]



from dataclasses import dataclass
from typing import Any, List, Optional

from omegaconf import DictConfig, OmegaConf

import autrainer
from autrainer.core.scripts.abstract_script import MockParser

from .abstract_preprocess_script import (
    AbstractPreprocessScript,
    PreprocessArgs,
)
from .utils import (
    add_hydra_args_to_sys,
    catch_cli_errors,
    run_hydra_cmd,
    running_in_notebook,
)


@dataclass
class PreprocessArgs(PreprocessArgs):
    num_workers: int
    pbar_frequency: int
    silent: bool


class PreprocessScript(AbstractPreprocessScript):
    def __init__(self) -> None:
        super().__init__(
            "preprocess",
            "Launch a data preprocessing configuration (Hydra).",
            extended_description=(
                "For more information on Hydra's command line line flags, see:\n"
                "https://hydra.cc/docs/advanced/hydra-command-line-flags/."
            ),
            epilog="Example: autrainer preprocess -cn config.yaml",
            dataclass=PreprocessArgs,
            unknown_args=True,
        )

    def add_arguments(self) -> None:
        super().add_arguments()
        self.parser.add_argument(
            "-n",
            "--num-workers",
            type=int,
            default=-1,
            metavar="N",
            required=False,
            help="Number of workers to use for preprocessing. Defaults to -1.",
        )
        self.parser.add_argument(
            "-p",
            "--pbar-frequency",
            type=int,
            default=100,
            metavar="P",
            required=False,
            help="Frequency of progress bar updates. Defaults to 100.",
        )
        self.parser.add_argument(
            "-s",
            "--silent",
            action="store_true",
            default=False,
            required=False,
            help="Disable progress bar output.",
        )

    def main(self, args: PreprocessArgs) -> None:
        import hydra

        self.num_workers = args.num_workers
        self.pbar_frequency = args.pbar_frequency
        self.silent = args.silent
        self._override_launcher(args)
        self.datasets = {}
        self.preprocessing = {}

        @autrainer.main("config")
        def main(cfg: DictConfig) -> None:
            from hydra.errors import MissingConfigException

            if self._id_in_dict(self.datasets, cfg.dataset.id):
                return

            self.datasets[cfg.dataset.id] = OmegaConf.to_container(cfg.dataset)
            preprocessing_cfg = None
            if cfg.dataset.get("features_subdir"):
                try:
                    preprocessing_cfg = OmegaConf.to_container(
                        hydra.compose(
                            f"preprocessing/{cfg.dataset.features_subdir}"
                        )["preprocessing"]
                    )
                except MissingConfigException:
                    pass
            self.preprocessing[cfg.dataset.id] = preprocessing_cfg

        main()
        self._preprocess_datasets()
        self._clean_up()

    def _preprocess_datasets(self) -> None:
        from concurrent.futures import ThreadPoolExecutor
        import os
        from pathlib import Path

        import pandas as pd
        from tqdm import tqdm

        from autrainer.datasets.utils import AbstractFileHandler
        from autrainer.transforms import SmartCompose

        def _get_csvs(path: str) -> List[str]:
            return [
                f
                for f in os.listdir(path)
                if os.path.isfile(os.path.join(path, f)) and f.endswith(".csv")
            ]

        def _read_csvs(path: str, files: List[str]) -> List[pd.DataFrame]:
            return [pd.read_csv(os.path.join(path, f)) for f in files]

        def _get_unique_files(
            dfs: List[pd.DataFrame],
            column: str,
        ) -> List[str]:
            return pd.concat(dfs, ignore_index=True)[column].unique().tolist()

        def _create_subdirs(path: str, files: List[str]) -> None:
            for f in files:
                os.makedirs(
                    os.path.dirname(os.path.join(path, f)),
                    exist_ok=True,
                )

        def _get_num_chunks() -> int:
            if self.num_workers == -1:
                return os.cpu_count() or 1
            return self.num_workers

        def _split_chunks(files: List[str], chunks: int) -> List[List[str]]:
            avg = len(files) / chunks
            out = []
            last = 0
            while last < len(files):
                out.append(files[int(last) : int(last + avg)])
                last += avg
            return out

        def _process_chunk(
            chunk: List[str],
            dataset: dict,
            preprocess: dict,
            pbar: tqdm,
            lock: Any,
        ) -> None:
            file_handler = autrainer.instantiate_shorthand(
                config=preprocess["file_handler"],
                instance_of=AbstractFileHandler,
            )
            output_file_handler = autrainer.instantiate_shorthand(
                config=dataset["file_handler"],
                instance_of=AbstractFileHandler,
            )
            pipeline = SmartCompose(
                [
                    autrainer.instantiate_shorthand(t)
                    for t in preprocess["pipeline"]
                ]
            )

            file_count = 0
            for file_path in chunk:
                in_path = Path(dataset["path"], "default", file_path)
                out_path = Path(
                    dataset["path"],
                    dataset["features_subdir"],
                    file_path,
                ).with_suffix("." + dataset["file_type"])
                data = file_handler.load(in_path)
                data = pipeline(data, 0)
                output_file_handler.save(out_path, data)
                file_count += 1
                if file_count % self.pbar_frequency == 0:
                    with lock:
                        pbar.update(self.pbar_frequency)
            with lock:
                pbar.update(file_count % self.pbar_frequency)

        print("Preprocessing datasets...")
        for (name, dataset), preprocess in zip(
            self.datasets.items(), self.preprocessing.values()
        ):
            print(f" - {name}")
            if preprocess is None or os.path.isdir(
                os.path.join(dataset["path"], dataset["features_subdir"])
            ):
                continue

            csvs = _get_csvs(dataset["path"])
            dfs = _read_csvs(dataset["path"], csvs)
            unique_files = _get_unique_files(dfs, dataset["index_column"])
            _create_subdirs(
                os.path.join(dataset["path"], dataset["features_subdir"]),
                unique_files,
            )
            num_chunks = _get_num_chunks()
            chunks = _split_chunks(unique_files, num_chunks)
            lock = tqdm.get_lock()
            with tqdm(
                total=len(unique_files), desc=name, disable=self.silent
            ) as pbar:
                with ThreadPoolExecutor(max_workers=num_chunks) as executor:
                    futures = [
                        executor.submit(
                            _process_chunk,
                            chunk,
                            dataset,
                            preprocess,
                            pbar,
                            lock,
                        )
                        for chunk in chunks
                    ]
                    for future in futures:
                        future.result()


@catch_cli_errors
def preprocess(
    override_kwargs: Optional[dict] = None,
    num_workers: int = -1,
    pbar_frequency: int = 100,
    silent: bool = False,
    cfg_launcher: bool = False,
    config_name: str = "config",
    config_path: Optional[str] = None,
) -> None:
    """Launch a data preprocessing configuration.

    Args:
        override_kwargs: Additional Hydra override arguments to pass to the
            train script.
        num_workers: Number of workers to use for preprocessing.
            Defaults to -1.
        pbar_frequency: Frequency of progress bar updates. Defaults to 100.
        silent: Disable progress bar output. Defaults to False.
        cfg_launcher: Use the launcher specified in the configuration instead
            of the Hydra basic launcher. Defaults to False.
        config_name: The name of the config (usually the file name without the
            .yaml extension). Defaults to "config".
        config_path: The config path, a directory where Hydra will search for
            config files. If config_path is None no directory is added to the
            search path. Defaults to None.
    """
    if running_in_notebook():
        cmd = "preprocess"
        if cfg_launcher:
            cmd += " -l"
        if silent:
            cmd += " -s"
        cmd += f" -n {num_workers} -p {pbar_frequency}"
        run_hydra_cmd(cmd, override_kwargs, config_name, config_path)

    else:
        add_hydra_args_to_sys(override_kwargs, config_name, config_path)
        script = PreprocessScript()
        script.parser = MockParser()
        script.main(
            PreprocessArgs(cfg_launcher, num_workers, pbar_frequency, silent)
        )



from dataclasses import dataclass
import os
import shutil

from omegaconf import OmegaConf

import autrainer
from autrainer.core.constants import CONFIG_FOLDERS

from .abstract_script import AbstractScript, MockParser
from .command_line_error import CommandLineError
from .utils import catch_cli_errors


@dataclass
class ShowArgs:
    directory: str
    config: str
    save: bool
    force: bool


class ShowScript(AbstractScript):
    def __init__(self) -> None:
        super().__init__(
            "show",
            "Show and save a global configuration.",
            epilog="Example: autrainer show model EfficientNet-B0 -s",
            dataclass=ShowArgs,
        )

    def add_arguments(self) -> None:
        self.parser.add_argument(
            "directory",
            type=str,
            help="The directory to list configurations from. Choose from:"
            + "\n - ".join([""] + CONFIG_FOLDERS),
        )
        self.parser.add_argument(
            "config",
            type=str,
            help=(
                "The global configuration to show. Configurations can be "
                "discovered using the 'autrainer list' command."
            ),
        )
        self.parser.add_argument(
            "-s",
            "--save",
            action="store_true",
            help="Save the global configuration to the local conf/ directory.",
        )
        self.parser.add_argument(
            "-f",
            "--force",
            action="store_true",
            help=(
                "Force overwrite local configuration if it exists "
                "in combination with -s/--save."
            ),
        )

    def main(self, args: ShowArgs) -> None:
        self._assert_config_directory(args.directory)
        path = self._get_path(args)
        self._assert_config_exists(path, args)
        self._show_configuration(path)
        self._save_configuration(path, args)

    def _get_path(self, args: ShowArgs) -> str:
        return os.path.join(
            os.path.dirname(autrainer.__path__[0]),
            "autrainer-configurations",
            args.directory,
            ShowScript._get_config_name(args.config),
        )

    @staticmethod
    def _get_config_name(config: str) -> str:
        if not config.endswith(".yaml"):
            return f"{config}.yaml"
        return config

    def _assert_config_exists(self, path: str, args: ShowArgs) -> None:
        if not os.path.exists(path):
            m = (
                f"No global configuration '{args.config}' found "
                f"for '{args.directory}'."
            )
            raise CommandLineError(self.parser, m, code=1)

    def _show_configuration(self, path: str) -> None:
        config = OmegaConf.load(path)
        print(OmegaConf.to_yaml(config))

    def _save_configuration(self, path: str, args: ShowArgs) -> None:
        if not args.save:
            return

        out_path = os.path.join(
            "conf",
            args.directory,
            ShowScript._get_config_name(args.config),
        )
        os.makedirs(os.path.dirname(out_path), exist_ok=True)
        if os.path.exists(out_path) and not args.force:
            m = (
                f"{args.directory} configuration '{args.config}' "
                "already exists. Use -f to overwrite."
            )
            raise CommandLineError(self.parser, m, code=1)
        shutil.copyfile(path, out_path)


@catch_cli_errors
def show(
    directory: str,
    config: str,
    save: bool = False,
    force: bool = False,
) -> None:
    """Show and save a global configuration.

    If called in a notebook, the function will not raise an error and print
    the error message instead.

    Args:
        directory: The directory to list configurations from. Choose from:
            :const:`~autrainer.core.constants.CONFIG_FOLDERS`.
        config: The global configuration to show. Configurations can be
            discovered using the 'autrainer list' command.
        save: Save the global configuration to the local conf directory.
            Defaults to False.
        force: Force overwrite local configuration if it exists in combination
            with save=True. Defaults to False.

    Raises:
        CommandLineError: If the global configuration does not exist.
        CommandLineError: If while saving the local configuration, the
            configuration already exists and force is not set.
    """
    script = ShowScript()
    script.parser = MockParser()
    script.main(ShowArgs(directory, config, save, force))
